import sys
import argparse
import os

import yaml

from contentctl.actions.detection_testing.GitHubService import (
    GithubService,
)
from contentctl.actions.validate import ValidateInputDto, Validate
from contentctl.actions.generate import (
    GenerateInputDto,
    DirectorOutputDto,
    Generate,
)
from contentctl.actions.reporting import ReportingInputDto, Reporting
from contentctl.actions.new_content import NewContentInputDto, NewContent
from contentctl.actions.doc_gen import DocGenInputDto, DocGen
from contentctl.actions.initialize import Initialize, InitializeInputDto
from contentctl.actions.deploy import Deploy, DeployInputDto
from contentctl.input.director import DirectorInputDto
from contentctl.objects.enums import (
    SecurityContentType,
    SecurityContentProduct,
    DetectionTestingMode,
    PostTestBehavior,
)
from contentctl.input.new_content_generator import NewContentGeneratorInputDto
from contentctl.helper.config_handler import ConfigHandler

from contentctl.objects.config import Config

from contentctl.objects.app import App
from contentctl.objects.test_config import TestConfig
from contentctl.actions.test import Test, TestInputDto, TestOutputDto


import tqdm
import functools


def configure_unattended(args: argparse.Namespace) -> argparse.Namespace:
    # disable all calls to tqdm - this is so that CI/CD contexts don't
    # have a large amount of output due to progress bar updates.
    tqdm.tqdm.__init__ = functools.partialmethod(
        tqdm.tqdm.__init__, disable=args.unattended
    )
    if args.unattended:
        if args.behavior != PostTestBehavior.never_pause.name:
            print(
                f"For unattended mode, --behavior MUST be {PostTestBehavior.never_pause.name}.\nUpdating the behavior from '{args.behavior}' to '{PostTestBehavior.never_pause.name}'"
            )
            args.behavior = PostTestBehavior.never_pause.name

    return args


def print_ascii_art():
    print(
        """
Running Splunk Security Content Control Tool (contentctl) 
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⢶⠛⡇⠀⠀⠀⠀⠀⠀⣠⣦⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⣀⠼⠖⠛⠋⠉⠉⠓⠢⣴⡻⣾⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⢀⡠⠔⠊⠁⠀⠀⠀⠀⠀⠀⣠⣤⣄⠻⠟⣏⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⣠⠞⠁⠀⠀⠀⡄⠀⠀⠀⠀⠀⠀⢻⣿⣿⠀⢀⠘⡄⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⢸⡇⠀⠀⠀⡠⠊⠀⠀⠀⠀⠀⠀⠀⠀⠉⠉⠀⠈⠁⠘⣄⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⢸⡉⠓⠒⠊⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠘⢄⠀⠀⠀⠈⢦⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠈⡇⠀⢠⠀⠀⠀⠀⠀⠀⠀⠈⡷⣄⠀⠀⢀⠈⠀⠀⠑⢄⠀⠑⢄⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠹⡄⠘⡄⠀⠀⠀⠀⢀⡠⠊⠀⠙⠀⠀⠈⢣⠀⠀⠀⢀⠀⠀⠀⠉⠒⠤⣀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠉⠁⠛⠲⢶⡒⠈⠁⠀⠀⠀⠀⠀⠀⠀⠀⠀⡄⠀⠀⠉⠂⠀⠀⠀⠀⠤⡙⠢⣄⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⢹⠀⠀⡀⠀⠀⢸⠀⠀⠀⠀⠘⠇⠀⠀⠀⠀⠀⠀⠀⠀⢀⠈⠀⠈⠳⡄⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠈⡇⠀⠣⠀⠀⠈⠀⢀⠀⠀⠀⠀⠀⠀⢀⣀⠀⠀⢀⡀⠀⠑⠄⠈⠣⡘⢆⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⢧⠀⠀⠀⠀⠀⠀⠿⠀⠀⠀⠀⣠⠞⠉⠀⠀⠀⠀⠙⢆⠀⠀⡀⠀⠁⠈⢇⠀
⠀⠀⠀⠀⠀⠀⠀⠀⢹⠀⢤⠀⠀⠀⠀⠀⠀⠀⠀⢰⠁⠀⠀⠀⠀⠀⠀⠀⠀⠁⠀⠙⡄⠀⡀⠈⡆
⠀⠀⠀⠀⠀⠀⠀⠀⠸⡆⠘⠃⠀⠀⠀⢀⡄⠀⠀⡇⠀⠀⡄⠀⠀⠀⠰⡀⠀⠀⡄⠀⠉⠀⠃⠀⢱
⠀⠀⠀⠀⠀⠀⠀⠀⠀⢣⡀⠀⠀⡆⠀⠸⠇⠀⠀⢳⠀⠀⠈⠀⠀⠀⠐⠓⠀⠀⢸⡄⠀⠀⠀⡀⢸
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢳⡀⠀⢻⠀⠀⠀⠀⢰⠛⢆⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠃⠀⡆⠀⠃⡼
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⣷⣤⣽⣧⠀⠀⠀⡜⠀⠈⠢⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⠃
⠀⠀⠀⠀⠀⠀⠀⠀⠀⣾⣿⣿⣇⡿⠹⣷⣄⣬⡗⠢⣤⠖⠛⢳⣤⣀⠀⠀⠀⠀⠀⠀⠀⠀⡰⠃⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠘⠋⢠⣾⢿⡏⣸⠀⠀⠈⠋⠛⠧⠤⠘⠛⠉⠙⠒⠒⠒⠒⠉⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠉⠻⠶⠃⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀

    By: Splunk Threat Research Team [STRT] - research@splunk.com
    """
    )


def start(args) -> Config:
    return ConfigHandler.read_config(os.path.join(args.path, "contentctl.yml"))


def initialize(args) -> None:

    Initialize().execute(InitializeInputDto(path=os.path.abspath(args.path)))


def build(args) -> DirectorOutputDto:
    config = start(args)
    product_type = SecurityContentProduct.splunk_app
    director_input_dto = DirectorInputDto(
        input_path=os.path.abspath(args.path), product=product_type, config=config
    )
    generate_input_dto = GenerateInputDto(director_input_dto)

    generate = Generate()

    return generate.execute(generate_input_dto)


def inspect(args) -> None:
    raise (Exception("WARNING - INSPECT NOT YET IMPLEMENTED"))
    # Inspect(args)


def deploy(args) -> None:
    config = start(args)
    deploy_input_dto = DeployInputDto(path=args.path, config=config)
    deploy = Deploy()
    deploy.execute(deploy_input_dto)


def test(args: argparse.Namespace):
    args = configure_unattended(args)
    config = start(args)

    # set some arguments that are not
    # yet exposed/written properly in
    # the config file
    test_config = TestConfig.parse_obj(
        {
            # "repo_path": args.path,
            "mode": args.mode,
            "num_containers": 1,
            "post_test_behavior": args.behavior,
            "detections_list": args.detections_list,
        }
    )

    # We do this before generating the app to save some time if options are incorrect.
    # For example, if the detection(s) we are trying to test do not exist
    githubService = GithubService(test_config)

    director_output_dto = build(args)

    # All this information will later come from the config, so we will
    # be able to do it in Test().execute. For now, we will do it here
    app = App(
        uid=9999,
        appid="my_custom_app",
        title="my_custom_app",
        release="1.0.0",
        http_path=None,
        local_path=os.path.join(
            os.path.abspath(args.path), f"{config.build.splunk_app.path}.tar.gz"
        ),
        description="some description",
        splunkbase_path=None,
    )

    # We need to do this instead of appending to retrigger validation.
    # It does not happen the first time since validation does not run for default values
    # unless we use always=True in the validator
    # we always want to keep CIM as the last app installed

    test_config.apps = [app] + test_config.apps

    test_input_dto = TestInputDto(
        director_output_dto=director_output_dto,
        githubService=githubService,
        config=test_config,
    )
    test = Test()

    try:
        result = test.execute(test_input_dto)
        # This return code is important.  Even if testing
        # fully completes, if everything does not pass then
        # we want to return a nonzero status code
        if result:
            sys.exit(0)
        else:
            sys.exit(1)

    except Exception as e:
        print("Error running contentctl test: {str(e)}")
        sys.exit(1)


def validate(args) -> None:
    config = start(args)
    product_type = SecurityContentProduct.splunk_app
    director_input_dto = DirectorInputDto(
        input_path=os.path.abspath(args.path), product=product_type, config=config
    )
    validate_input_dto = ValidateInputDto(director_input_dto=director_input_dto)
    validate = Validate()
    return validate.execute(validate_input_dto)


def doc_gen(args) -> None:
    config = start(args)
    director_input_dto = DirectorInputDto(
        input_path=args.path, product=SecurityContentProduct.splunk_app, config=config
    )

    doc_gen_input_dto = DocGenInputDto(director_input_dto=director_input_dto)

    doc_gen = DocGen()
    doc_gen.execute(doc_gen_input_dto)


def new_content(args) -> None:

    if args.type == "detection":
        contentType = SecurityContentType.detections
    elif args.type == "story":
        contentType = SecurityContentType.stories
    else:
        print("ERROR: type " + args.type + " not supported")
        sys.exit(1)

    new_content_generator_input_dto = NewContentGeneratorInputDto(type=contentType)
    new_content_input_dto = NewContentInputDto(
        new_content_generator_input_dto, os.path.abspath(args.path)
    )
    new_content = NewContent()
    new_content.execute(new_content_input_dto)


def reporting(args) -> None:
    config = start(args)
    director_input_dto = DirectorInputDto(
        input_path=args.path, product=SecurityContentProduct.splunk_app, config=config
    )

    reporting_input_dto = ReportingInputDto(director_input_dto=director_input_dto)

    reporting = Reporting()
    reporting.execute(reporting_input_dto)


def main():
    """
    main function parses the arguments passed to the script and calls the respctive method.
    :param args: arguments passed by the user on command line while calling the script.
    :return: returns the output of the function called.
    """

    # grab arguments
    parser = argparse.ArgumentParser(
        description="Use `contentctl action -h` to get help with any Splunk content action"
    )
    parser.add_argument(
        "-p",
        "--path",
        required=False,
        default=".",
        help="path to the content path containing the contentctl.yml",
    )

    parser.set_defaults(func=lambda _: parser.print_help())
    actions_parser = parser.add_subparsers(
        title="Splunk content actions", dest="action"
    )

    # available actions
    init_parser = actions_parser.add_parser(
        "init",
        help="initialize a Splunk content pack using and customizes a configuration under contentctl.yml",
    )
    validate_parser = actions_parser.add_parser(
        "validate", help="validates a Splunk content pack"
    )
    build_parser = actions_parser.add_parser(
        "build", help="builds a Splunk content pack package to be distributed"
    )
    new_content_parser = actions_parser.add_parser(
        "new", help="create new Splunk content object (detection, or story)"
    )
    reporting_parser = actions_parser.add_parser(
        "report", help="create Splunk content report of the current pack"
    )
    inspect_parser = actions_parser.add_parser(
        "inspect",
        help="runs Splunk appinspect on a build Splunk app to ensure that an app meets Splunkbase requirements.",
    )
    deploy_parser = actions_parser.add_parser(
        "deploy", help="install an application on a target Splunk instance."
    )
    docs_parser = actions_parser.add_parser(
        "docs", help="create documentation in docs folder"
    )

    test_parser = actions_parser.add_parser(
        "test",
        help="Run a test of the detections against a Splunk Server or Splunk Docker Container",
    )

    init_parser.set_defaults(func=initialize)

    validate_parser.set_defaults(func=validate)

    build_parser.set_defaults(func=build)

    docs_parser.set_defaults(func=doc_gen)

    new_content_parser.add_argument(
        "-t",
        "--type",
        required=True,
        type=str,
        help="Type of security content object, choose between `detection`, `story`",
    )
    new_content_parser.set_defaults(func=new_content)

    reporting_parser.set_defaults(func=reporting)

    inspect_parser.add_argument(
        "-ap",
        "--app_path",
        required=False,
        type=str,
        default=None,
        help="path to the Splunk app to be inspected",
    )
    inspect_parser.set_defaults(func=inspect)

    deploy_parser.set_defaults(func=deploy)

    test_parser.add_argument(
        "--mode",
        required=False,
        default=DetectionTestingMode.all.name,
        type=str,
        choices=DetectionTestingMode._member_names_,
        help="Controls which detections to test. 'all' will test all detections in the repo."
        "'selected' will test a list of detections that have "
        "been provided via the --selected command line argument (see for more details).",
    )
    test_parser.add_argument(
        "--behavior",
        required=False,
        default=PostTestBehavior.pause_on_failure.name,
        type=str,
        choices=PostTestBehavior._member_names_,
        help="Controls what to do when a test completes. 'always_pause' means that the state of "
        "the test will always pause after a test, allowing the user to log into the "
        "server and experiment with the search and data before it is removed.  'pause_on_failure' "
        "will pause execution ONLY when a test fails. The user may press ENTER in the terminal "
        "running the test to move on to the next test.  'never_pause' will never stop testing, "
        "even if a test fails. Please note that 'never_pause' MUST be used for a test to "
        "run in an unattended manner or in a CI/CD system - otherwise a single failed test "
        "will result in the testing never finishing as the tool waits for input.",
    )
    test_parser.add_argument(
        "-d",
        "--detections_list",
        required=False,
        nargs="+",
        type=str,
        help="An explicit list "
        "of detections to test. Their paths should be relative to the app path.",
    )

    test_parser.add_argument("--unattended", action=argparse.BooleanOptionalAction)

    test_parser.set_defaults(func=test)

    # parse them
    args = parser.parse_args()

    print_ascii_art()
    args.func(args)



from pydantic import BaseModel, PrivateAttr, Field
from dataclasses import dataclass
import abc
import requests
import splunklib.client as client
from contentctl.objects.enums import PostTestBehavior
from contentctl.objects.detection import Detection
from contentctl.objects.unit_test_test import UnitTestTest
from contentctl.objects.unit_test_attack_data import UnitTestAttackData
from contentctl.objects.unit_test_result import UnitTestResult
from contentctl.objects.test_config import TestConfig
from shutil import copyfile
from splunklib.binding import HTTPError
import os.path
import configparser
from ssl import SSLEOFError, SSLZeroReturnError
import time
import uuid


from tempfile import TemporaryDirectory, mktemp
import pathlib
from contentctl.helper.utils import Utils
from contentctl.actions.detection_testing.DataManipulation import (
    DataManipulation,
)
import splunklib.results
from urllib3 import disable_warnings
import urllib.parse
import json
from typing import Union
import datetime
import tqdm


MAX_TEST_NAME_LENGTH = 70
TESTING_STATES = [
    "Downloading Data",
    "Replaying Data",
    "Waiting for Processing",
    "Running Search",
    "Deleting Data",
]

LONGEST_STATE = max(len(w) for w in TESTING_STATES)
PBAR_FORMAT_STRING = "{test_name} >> {state} | Time: {time}"


class ContainerStoppedException(Exception):
    pass


@dataclass(frozen=False)
class DetectionTestingManagerOutputDto:
    inputQueue: list[Detection] = Field(default_factory=list)
    outputQueue: list[Detection] = Field(default_factory=list)
    currentTestingQueue: dict[str, Union[Detection, None]] = Field(default_factory=dict)
    start_time: Union[datetime.datetime, None] = None
    replay_index: str = "CONTENTCTL_TESTING_INDEX"
    replay_host: str = "CONTENTCTL_HOST"
    timeout_seconds: int = 120
    terminate: bool = False


class DetectionTestingInfrastructure(BaseModel, abc.ABC):
    # thread: threading.Thread = threading.Thread()
    config: TestConfig
    sync_obj: DetectionTestingManagerOutputDto
    hec_token: str = ""
    hec_channel: str = ""
    _conn: client.Service = PrivateAttr()
    pbar: tqdm.tqdm = None
    start_time: float = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)

    def start(self):
        raise (
            NotImplementedError(
                "start() is not implemented for Abstract Type DetectionTestingInfrastructure"
            )
        )

    def get_name(self) -> str:
        raise (
            NotImplementedError(
                "get_name() is not implemented for Abstract Type DetectionTestingInfrastructure"
            )
        )

    def setup(self):
        self.pbar = tqdm.tqdm(
            total=100,
            initial=0,
            bar_format=f"{self.get_name()} starting",
            miniters=0,
            mininterval=0,
        )
        self.start_time = time.time()
        try:
            for func, msg in [
                (self.start, "Starting"),
                (self.get_conn, "Getting API Connection"),
                (
                    self.create_replay_index,
                    f"Create index '{self.sync_obj.replay_index}'",
                ),
                (self.configure_imported_roles, "Configuring Roles"),
                (self.configure_delete_indexes, "Configuring Indexes"),
                (self.configure_conf_file_datamodels, "Configuring Datamodels"),
                (self.configure_hec, "Configuring HEC"),
                (self.wait_for_ui_ready, "Waiting for UI"),
            ]:

                self.format_pbar_string(
                    self.get_name(), msg, self.start_time, update_sync_status=True
                )
                func()
                self.check_for_teardown()

        except Exception as e:
            self.pbar.write(str(e))
            self.finish()
            return

        self.format_pbar_string(self.get_name(), "Finished Setup!", self.start_time)

    def wait_for_ui_ready(self):
        self.get_conn()

    def configure_hec(self):
        self.hec_channel = str(uuid.uuid4())
        try:
            res = self.get_conn().input(
                path="/servicesNS/nobody/splunk_httpinput/data/inputs/http/http:%2F%2FDETECTION_TESTING_HEC"
            )
            self.hec_token = str(res.token)
            return
        except Exception as e:
            # HEC input does not exist.  That's okay, we will create it
            pass

        try:

            res = self.get_conn().inputs.create(
                name="DETECTION_TESTING_HEC",
                kind="http",
                index=self.sync_obj.replay_index,
                indexes=f"{self.sync_obj.replay_index},_internal,_audit",
                useACK=True,
            )
            self.hec_token = str(res.token)
            return

        except Exception as e:
            raise (Exception(f"Failure creating HEC Endpoint: {str(e)}"))

    def get_conn(self) -> client.Service:
        try:
            if not self._conn:
                self.connect_to_api()
            elif self._conn.restart_required:
                # continue trying to re-establish a connection until after
                # the server has restarted
                self.connect_to_api()
        except Exception as e:
            # there was some issue getting the connection. Try again just once
            self.connect_to_api()
        return self._conn

    def check_for_teardown(self):
        # Make sure we can easily quit during setup if we need to.
        # Some of these stages can take a long time
        if self.sync_obj.terminate:
            # Exiting in a thread just quits the thread, not the entire process
            raise (ContainerStoppedException(f"Testing stopped for {self.get_name()}"))

    def connect_to_api(self, sleep_seconds: int = 5):
        while True:
            self.check_for_teardown()
            try:

                conn = client.connect(
                    host=self.config.test_instance_address,
                    port=self.config.api_port,
                    username=self.config.splunk_app_username,
                    password=self.config.splunk_app_password,
                )

                if conn.restart_required:
                    self.format_pbar_string(
                        self.get_name(),
                        "Waiting for reboot",
                        self.start_time,
                        update_sync_status=True,
                    )
                else:
                    # Finished setup
                    self._conn = conn
                    return

            except ConnectionRefusedError as e:
                raise (e)
            except SSLEOFError as e:
                pass
            except SSLZeroReturnError as e:
                pass
            except ConnectionResetError:
                pass
            except Exception as e:
                self.pbar.write(
                    f"Error getting API connection (not quitting) '{type(e).__name__}': {str(e)}"
                )
                # self.pbar.write(
                #     f"Unhandled exception getting connection to splunk server: {str(e)}"
                # )
                # self.sync_obj.terminate = True

            for _ in range(sleep_seconds):
                self.format_pbar_string(
                    self.get_name(),
                    "Getting API Connection",
                    self.start_time,
                    update_sync_status=True,
                )
                time.sleep(1)

    def create_replay_index(self):

        try:
            self.get_conn().indexes.create(name=self.sync_obj.replay_index)
        except HTTPError as e:
            if b"already exists" in e.body:
                pass
            else:
                raise Exception(
                    f"Error creating index {self.sync_obj.replay_index} - {str(e)}"
                )

    def configure_imported_roles(
        self,
        imported_roles: list[str] = ["user", "power", "can_delete"],
        indexes: list[str] = ["_*", "*"],
    ):
        indexes.append(self.sync_obj.replay_index)
        indexes_encoded = ";".join(indexes)
        self.get_conn().roles.post(
            self.config.splunk_app_username,
            imported_roles=imported_roles,
            srchIndexesAllowed=indexes_encoded,
            srchIndexesDefault=self.sync_obj.replay_index,
        )

    def configure_delete_indexes(self, indexes: list[str] = ["_*", "*"]):
        indexes.append(self.sync_obj.replay_index)
        endpoint = "/services/properties/authorize/default/deleteIndexesAllowed"
        indexes_encoded = ";".join(indexes)
        try:
            self.get_conn().post(endpoint, value=indexes_encoded)
        except Exception as e:
            self.pbar.write(
                f"Error configuring deleteIndexesAllowed with '{indexes_encoded}': [{str(e)}]"
            )

    def wait_for_conf_file(self, app_name: str, conf_file_name: str):
        while True:
            self.check_for_teardown()
            time.sleep(1)
            try:
                res = self.get_conn().get(
                    f"configs/conf-{conf_file_name}", app=app_name
                )
                return
            except Exception as e:
                pass
                self.format_pbar_string(
                    self.get_name(), "Configuring Datamodels", self.start_time
                )

    def configure_conf_file_datamodels(self, APP_NAME: str = "Splunk_SA_CIM"):
        self.wait_for_conf_file(APP_NAME, "datamodels")

        parser = configparser.ConfigParser()
        cim_acceleration_datamodels = pathlib.Path(
            os.path.join(
                os.path.dirname(__file__), "../../../templates/datamodels_cim.conf"
            )
        )

        custom_acceleration_datamodels = pathlib.Path(
            os.path.join(
                os.path.dirname(__file__), "../../../templates/datamodels_custom.conf"
            )
        )

        if custom_acceleration_datamodels.is_file():
            parser.read(custom_acceleration_datamodels)
            if len(parser.keys()) > 1:
                self.pbar.write(
                    f"Read {len(parser)-1} custom datamodels from {str(custom_acceleration_datamodels)}!"
                )

        if not cim_acceleration_datamodels.is_file():
            self.pbar.write(
                f"******************************\nDATAMODEL ACCELERATION FILE {str(cim_acceleration_datamodels)} NOT FOUND. CIM DATAMODELS NOT ACCELERATED\n******************************\n"
            )
        else:
            parser.read(cim_acceleration_datamodels)

        for datamodel_name in parser:
            if datamodel_name == "DEFAULT":
                # Skip the DEFAULT section for configparser
                continue
            for name, value in parser[datamodel_name].items():
                try:
                    res = self.get_conn().post(
                        f"properties/datamodels/{datamodel_name}/{name}",
                        app=APP_NAME,
                        value=value,
                    )

                except Exception as e:
                    self.pbar.write(
                        f"Error creating the conf Datamodel {datamodel_name} key/value {name}/{value}: {str(e)}"
                    )

    def execute(self):
        while True:
            try:
                self.check_for_teardown()
            except ContainerStoppedException as e:
                self.finish()
                return

            try:
                detection = self.sync_obj.inputQueue.pop()
                self.sync_obj.currentTestingQueue[self.get_name()] = detection
            except IndexError as e:
                # self.pbar.write(
                #    f"No more detections to test, shutting down {self.get_name()}"
                # )
                self.finish()
                return
            try:
                self.test_detection(detection)
            except ContainerStoppedException as e:
                self.pbar.write(f"Stopped container [{self.get_name()}]")
                self.finish()
                return
            except Exception as e:
                self.pbar.write(f"Error testing detection: {str(e)}")
            finally:
                self.sync_obj.outputQueue.append(detection)
                self.sync_obj.currentTestingQueue[self.get_name()] = None

    def test_detection(self, detection: Detection):
        if detection.test is None:
            self.pbar.write(f"No test(s) found for {detection.name}")
            return

        for test in detection.test.tests:
            self.execute_test(detection, test)

    def format_pbar_string(
        self,
        test_name: str,
        state: str,
        start_time: Union[float, None],
        set_pbar: bool = True,
        update_sync_status: bool = False,
    ) -> str:
        if start_time == None:
            start_time = self.start_time
        field_one = test_name.ljust(MAX_TEST_NAME_LENGTH)
        field_two = state.ljust(LONGEST_STATE)
        field_three = datetime.timedelta(seconds=round(time.time() - start_time))
        new_string = PBAR_FORMAT_STRING.format(
            test_name=field_one, state=field_two, time=field_three
        )
        if set_pbar:
            self.pbar.bar_format = new_string
            self.pbar.update()
        if update_sync_status:
            self.sync_obj.currentTestingQueue[self.get_name()] = {
                "name": state,
                "search": "N/A",
            }
        return new_string

    def execute_test(
        self, detection: Detection, test: UnitTestTest, FORCE_ALL_TIME: bool = True
    ):
        start_time = time.time()
        self.pbar.reset()
        self.format_pbar_string(test.name, "Beginning Test", start_time)
        # https://github.com/WoLpH/python-progressbar/issues/164
        # Use NullBar if there is more than 1 container or we are running
        # in a non-interactive context

        try:
            self.replay_attack_data_files(test.attack_data, test, start_time)
        except Exception as e:

            test.result = UnitTestResult()
            test.result.set_job_content(
                e, self.config, duration=time.time() - start_time
            )
            self.pbar.write(
                self.format_pbar_string(
                    test.name,
                    "\x1b[0;30;41m" + "FAIL".ljust(LONGEST_STATE) + "\x1b[0m",
                    start_time=time.time() - start_time,
                    set_pbar=False,
                )
            )

            return

        # Set the mode and timeframe, if required
        kwargs = {"exec_mode": "blocking"}
        for baseline in test.baselines:

            self.retry_search_until_timeout(detection, test, kwargs, start_time)

        if not FORCE_ALL_TIME:
            if test.earliest_time is not None:
                kwargs.update({"earliest_time": test.earliest_time})
            if test.latest_time is not None:
                kwargs.update({"latest_time": test.latest_time})

        try:
            self.retry_search_until_timeout(detection, test, kwargs, start_time)
        except ContainerStoppedException as e:
            raise (e)
        except Exception as e:
            test.result = UnitTestResult()
            test.result.set_job_content(
                e, self.config, duration=time.time() - start_time
            )

        if (
            self.config.post_test_behavior == PostTestBehavior.always_pause
            or (
                self.config.post_test_behavior == PostTestBehavior.pause_on_failure
                and (test.result is None or test.result.success == False)
            )
        ) and not self.sync_obj.terminate:
            if test.result is None:
                res = "ERROR"
                link = detection.search
            else:
                res = test.result.success
                if res:
                    res = "PASS"
                else:
                    res = "FAIL"
                link = test.result.get_summary_dict()["sid_link"]

            self.format_pbar_string(
                test.name, f"{res} - {link} (CTRL+D to continue)", start_time
            )

            try:
                _ = input()
            except Exception as e:
                pass

        self.format_pbar_string(test.name, f"Deleting Data", start_time)
        self.delete_attack_data(test.attack_data)

        if test.result is not None and test.result.success:
            self.pbar.write(
                self.format_pbar_string(
                    test.name,
                    "\x1b[0;30;42m" + "PASS".ljust(LONGEST_STATE) + "\x1b[0m",
                    start_time,
                    set_pbar=False,
                )
            )

        else:
            self.pbar.write(
                self.format_pbar_string(
                    test.name,
                    "\x1b[0;30;41m" + "FAIL".ljust(LONGEST_STATE) + "\x1b[0m",
                    start_time,
                    set_pbar=False,
                )
            )

        if test.result is not None:
            test.result.duration = round(time.time() - start_time, 2)

    def retry_search_until_timeout(
        self,
        detection: Detection,
        test: UnitTestTest,
        kwargs: dict,
        start_time: float,
    ):
        search_start_time = time.time()
        search_stop_time = time.time() + self.sync_obj.timeout_seconds

        if test.pass_condition is None:
            # we will default to ensuring at least one result exists
            search = detection.search
        else:
            search = f"{detection.search} {test.pass_condition}"

        # Search that do not begin with '|' must begin with 'search '
        if not search.strip().startswith("|"):
            if not search.strip().startswith("search "):
                search = f"search {search}"

        # exponential backoff for wait time
        tick = 2

        while time.time() < search_stop_time:

            for _ in range(pow(2, tick - 1)):
                # This loop allows us to capture shutdown events without being
                # stuck in an extended sleep. Remember that this raises an exception
                self.check_for_teardown()
                self.format_pbar_string(test.name, "Waiting for Processing", start_time)

                time.sleep(1)

            self.format_pbar_string(test.name, "Running Search", start_time)

            job = self.get_conn().search(query=search, **kwargs)

            # the following raises an error if there is an exception in the search
            _ = job.results(output_mode="json")

            if int(job.content.get("resultCount", "0")) > 0:
                test.result = UnitTestResult()
                test.result.set_job_content(
                    job.content,
                    self.config,
                    success=True,
                    duration=time.time() - search_start_time,
                )

                return
            else:
                test.result = UnitTestResult()
                test.result.set_job_content(
                    job.content,
                    self.config,
                    success=False,
                    duration=time.time() - search_start_time,
                )

            tick += 1

        return

    def delete_attack_data(self, attack_data_files: list[UnitTestAttackData]):
        for attack_data_file in attack_data_files:
            index = attack_data_file.custom_index or self.sync_obj.replay_index
            host = attack_data_file.host or self.sync_obj.replay_host
            splunk_search = f'search index="{index}" host="{host}" | delete'
            kwargs = {"exec_mode": "blocking"}
            try:

                job = self.get_conn().jobs.create(splunk_search, **kwargs)
                results_stream = job.results(output_mode="json")
                reader = splunklib.results.JSONResultsReader(results_stream)

            except Exception as e:
                raise (
                    Exception(
                        f"Trouble deleting data using the search {splunk_search}: {str(e)}"
                    )
                )

    def replay_attack_data_files(
        self,
        attack_data_files: list[UnitTestAttackData],
        test: UnitTestTest,
        start_time: float,
    ):
        with TemporaryDirectory(prefix="contentctl_attack_data") as attack_data_dir:
            for attack_data_file in attack_data_files:
                self.replay_attack_data_file(
                    attack_data_file, attack_data_dir, test, start_time
                )

    def replay_attack_data_file(
        self,
        attack_data_file: UnitTestAttackData,
        tmp_dir: str,
        test: UnitTestTest,
        start_time: float,
    ):
        tempfile = mktemp(dir=tmp_dir)

        if not (
            attack_data_file.data.startswith("https://")
            or attack_data_file.data.startswith("http://")
        ):
            if pathlib.Path(attack_data_file.data).is_file():
                self.format_pbar_string(test.name, "Copying Data", start_time)
                try:
                    copyfile(attack_data_file.data, tempfile)
                except Exception as e:
                    raise (
                        Exception(
                            f"Error copying local Attack Data File for [{Detection.name}] - [{attack_data_file.data}]: {str(e)}"
                        )
                    )
            else:
                raise (
                    Exception(
                        f"Attack Data File for [{Detection.name}] is local [{attack_data_file.data}], but does not exist."
                    )
                )

        else:
            # Download the file
            # We need to overwrite the file - mkstemp will create an empty file with the
            # given name
            try:
                # In case the path is a local file, try to get it

                self.format_pbar_string(test.name, "Downloading Data", start_time)

                Utils.download_file_from_http(
                    attack_data_file.data, tempfile, self.pbar, overwrite_file=True
                )
            except Exception as e:
                raise (
                    Exception(
                        f"Could not download attack data file [{attack_data_file.data}]:{str(e)}"
                    )
                )

        # Update timestamps before replay
        if attack_data_file.update_timestamp:
            data_manipulation = DataManipulation()
            data_manipulation.manipulate_timestamp(
                tempfile, attack_data_file.sourcetype, attack_data_file.source
            )

        # Upload the data
        self.format_pbar_string(test.name, "Replaying Data", start_time)

        self.hec_raw_replay(tempfile, attack_data_file)

        return attack_data_file.custom_index or self.sync_obj.replay_index

    def hec_raw_replay(
        self,
        tempfile: str,
        attack_data_file: UnitTestAttackData,
        verify_ssl: bool = False,
    ):
        if verify_ssl is False:
            # need this, otherwise every request made with the requests module
            # and verify=False will print an error to the command line
            disable_warnings()

        # build the headers

        headers = {
            "Authorization": f"Splunk {self.hec_token}",  # token must begin with 'Splunk '
            "X-Splunk-Request-Channel": self.hec_channel,
        }

        url_params = {
            "index": attack_data_file.custom_index or self.sync_obj.replay_index,
            "source": attack_data_file.source,
            "sourcetype": attack_data_file.sourcetype,
            "host": attack_data_file.host or self.sync_obj.replay_host,
        }

        if self.config.test_instance_address.strip().lower().startswith("https://"):
            address_with_scheme = self.config.test_instance_address.strip().lower()
        elif self.config.test_instance_address.strip().lower().startswith("http://"):
            address_with_scheme = (
                self.config.test_instance_address.strip()
                .lower()
                .replace("http://", "https://")
            )
        else:
            address_with_scheme = f"https://{self.config.test_instance_address}"

        # Generate the full URL, including the host, the path, and the params.
        # We can be a lot smarter about this (and pulling the port from the url, checking
        # for trailing /, etc, but we leave that for the future)
        url_with_port = f"{address_with_scheme}:{self.config.hec_port}"
        url_with_hec_path = urllib.parse.urljoin(
            url_with_port, "services/collector/raw"
        )
        with open(tempfile, "rb") as datafile:
            try:
                res = requests.post(
                    url_with_hec_path,
                    params=url_params,
                    data=datafile.read(),
                    allow_redirects=True,
                    headers=headers,
                    verify=verify_ssl,
                )
                jsonResponse = json.loads(res.text)

            except Exception as e:
                raise (
                    Exception(
                        f"There was an exception sending attack_data to HEC: {str(e)}"
                    )
                )

        if "ackId" not in jsonResponse:
            raise (
                Exception(
                    f"key 'ackID' not present in response from HEC server: {jsonResponse}"
                )
            )

        ackId = jsonResponse["ackId"]
        url_with_hec_ack_path = urllib.parse.urljoin(
            url_with_port, "services/collector/ack"
        )

        requested_acks = {"acks": [jsonResponse["ackId"]]}
        while True:
            try:

                res = requests.post(
                    url_with_hec_ack_path,
                    json=requested_acks,
                    allow_redirects=True,
                    headers=headers,
                    verify=verify_ssl,
                )

                jsonResponse = json.loads(res.text)

                if "acks" in jsonResponse and str(ackId) in jsonResponse["acks"]:
                    if jsonResponse["acks"][str(ackId)] is True:
                        # ackID has been found for our request, we can return as the data has been replayed
                        return
                    else:
                        # ackID is not yet true, we will wait some more
                        time.sleep(2)

                else:
                    raise (
                        Exception(
                            f"Proper ackID structure not found for ackID {ackId} in {jsonResponse}"
                        )
                    )
            except Exception as e:
                raise (Exception(f"There was an exception in the post: {str(e)}"))

    def status(self):
        pass

    def finish(self):
        self.pbar.bar_format = f"Stopped container [{self.get_name()}]"
        self.pbar.update()
        self.pbar.close()

    def check_health(self):
        pass




from pydantic import BaseModel, validator, ValidationError


class PlaybookTag(BaseModel):
    analytic_story: list = None
    detections: list = None
    platform_tags: list = None
    playbook_fields: list = None
    product: list = None
    playbook_fields: list = None
    detection_objects: list = None  
    


from pydantic import BaseModel, validator, ValidationError


from contentctl.objects.unit_test_attack_data import UnitTestAttackData
from contentctl.objects.unit_test_baseline import UnitTestBaseline
from contentctl.objects.unit_test_result import UnitTestResult
from typing import Union


class UnitTestTest(BaseModel):
    name: str
    file: str
    pass_condition: Union[str, None] = None
    earliest_time: Union[str, None] = None
    latest_time: Union[str, None] = None
    baselines: list[UnitTestBaseline] = []
    attack_data: list[UnitTestAttackData]
    result: Union[None, UnitTestResult]



import sys

from pydantic import ValidationError

from contentctl.input.yml_reader import YmlReader
from contentctl.objects.baseline import Baseline
from contentctl.objects.enums import SecurityContentType
from contentctl.objects.enums import SecurityContentProduct


class BaselineBuilder():
    baseline : Baseline

    def setObject(self, path: str) -> None:
        yml_dict = YmlReader.load_file(path)
        yml_dict["tags"]["name"] = yml_dict["name"]

        try:
            self.baseline = Baseline.parse_obj(yml_dict)
        
        except ValidationError as e:
            print('Validation Error for file ' + path)
            print(e)
            sys.exit(1)

    def addDeployment(self, deployments: list) -> None:
        matched_deployments = []

        for d in deployments:
            d_tags = dict(d.tags)
            for d_tag in d_tags.keys():
                for attr in dir(self.baseline):
                    if not (attr.startswith('__') or attr.startswith('_')):
                        if attr == d_tag:
                            if type(self.baseline.__getattribute__(attr)) is str:
                                attr_values = [self.baseline.__getattribute__(attr)]
                            else:
                                attr_values = self.baseline.__getattribute__(attr)
                            
                            for attr_value in attr_values:
                                if attr_value == d_tags[d_tag]:
                                    matched_deployments.append(d)

        if len(matched_deployments) == 0:
            raise ValueError('No deployment found for baseline: ' + self.baseline.name)

        self.baseline.deployment = matched_deployments[-1]


    def reset(self) -> None:
        self.baseline = None


    def getObject(self) -> Baseline:
        return self.baseline



from pycvesearch import CVESearch

CVESSEARCH_API_URL = 'https://cve.circl.lu'


class CveEnrichment():

    @classmethod
    def enrich_cve(self, cve_id: str) -> dict:
        cve_enriched = dict()
        try:
            cve = CVESearch(CVESSEARCH_API_URL)
            result = cve.id(cve_id)
            cve_enriched['id'] = cve_id
            cve_enriched['cvss'] = result['cvss']
            cve_enriched['summary'] = result['summary']
        except Exception as e:
            # there was a error calling the circl api lets just empty the object
            print("WARNING, issue enriching {0}, with error: {1}".format(cve_id, str(e)))
            cve_enriched = dict()
    
        return cve_enriched




import sys

from dataclasses import dataclass

from pydantic import ValidationError
from typing import Union

from contentctl.objects.enums import SecurityContentProduct
from contentctl.input.director import (
    Director,
    DirectorInputDto,
    DirectorOutputDto,
)


@dataclass(frozen=True)
class ValidateInputDto:
    director_input_dto: DirectorInputDto


class Validate:
    def execute(self, input_dto: ValidateInputDto) -> None:
        director_output_dto = DirectorOutputDto([], [], [], [], [], [], [], [])
        director = Director(director_output_dto)
        director.execute(input_dto.director_input_dto)

        # uuid validation all objects
        try:
            security_content_objects = (
                director_output_dto.detections
                + director_output_dto.stories
                + director_output_dto.baselines
                + director_output_dto.investigations
                + director_output_dto.playbooks
            )
            self.validate_duplicate_uuids(security_content_objects)

            # validate tests
            self.validate_detection_exist_for_test(
                director_output_dto.tests, director_output_dto.detections
            )

        except ValueError as e:
            print(e)
            sys.exit(1)

        return None

    def validate_duplicate_uuids(self, security_content_objects):
        all_uuids = set()
        duplicate_uuids = set()
        for elem in security_content_objects:
            if elem.id in all_uuids:
                # The uuid has been found more than once
                duplicate_uuids.add(elem.id)
            else:
                # This is the first time the uuid has been found
                all_uuids.add(elem.id)

        if len(duplicate_uuids) == 0:
            return

        # At least once duplicate uuid has been found. Enumerate all
        # the pieces of content that use duplicate uuids
        content_with_duplicate_uuid = [
            content_object
            for content_object in security_content_objects
            if content_object.id in duplicate_uuids
        ]

        raise ValueError(
            "ERROR: Duplicate ID found in objects:\n"
            + "\n".join([obj.name for obj in content_with_duplicate_uuid])
        )

    def validate_detection_exist_for_test(self, tests: list, detections: list):
        for test in tests:
            found_detection = False
            for detection in detections:
                if test.tests[0].file in detection.file_path:
                    found_detection = True

            if not found_detection:
                raise ValueError(
                    "ERROR: detection doesn't exist for test file: " + test.name
                )



import re

from pydantic import BaseModel, validator, ValidationError
from contentctl.objects.mitre_attack_enrichment import MitreAttackEnrichment
from contentctl.objects.constants import *

class DetectionTags(BaseModel):
    # detection spec
    name: str
    analytic_story: list
    asset_type: str
    automated_detection_testing: str = None
    cis20: list = None
    confidence: str
    context: list
    dataset: list = None
    impact: int
    kill_chain_phases: list
    message: str
    mitre_attack_id: list = None
    nist: list = None
    observable: list
    product: list
    required_fields: list
    risk_score: int
    security_domain: str
    risk_severity: str = None
    cve: list = None
    supported_tas: list = None

    # enrichment
    mitre_attack_enrichments: list[MitreAttackEnrichment] = []
    confidence_id: int = None
    impact_id: int = None
    context_ids: list = None
    risk_level_id: int = None
    risk_level: str = None
    observable_str: str = None
    kill_chain_phases_id: list = None


    @validator('cis20')
    def tags_cis20(cls, v, values):
        pattern = 'CIS [0-9]{1,2}'
        for value in v:
            if not re.match(pattern, value):
                raise ValueError('CIS controls are not following the pattern CIS xx: ' + values["name"])
        return v

    @validator('confidence')
    def tags_confidence(cls, v, values):
        v = int(v)
        if not (v > 0 and v <= 100):
             raise ValueError('confidence score is out of range 1-100: ' + values["name"])
        else:
            return v

    @validator('context')
    def tags_context(cls, v, values):
        context_list = SES_CONTEXT_MAPPING.keys()
        for value in v:
            if value not in context_list:
                raise ValueError('context value not valid for ' + values["name"] + '. valid options are ' + str(context_list) )
        return v

    @validator('impact')
    def tags_impact(cls, v, values):
        if not (v > 0 and v <= 100):
             raise ValueError('impact score is out of range 1-100: ' + values["name"])
        else:
            return v

    @validator('kill_chain_phases')
    def tags_kill_chain_phases(cls, v, values):
        valid_kill_chain_phases = SES_KILL_CHAIN_MAPPINGS.keys()
        for value in v:
            if value not in valid_kill_chain_phases:
                raise ValueError('kill chain phase not valid for ' + values["name"] + '. valid options are ' + str(valid_kill_chain_phases))
        return v

    @validator('mitre_attack_id')
    def tags_mitre_attack_id(cls, v, values):
        pattern = 'T[0-9]{4}'
        for value in v:
            if not re.match(pattern, value):
                raise ValueError('Mitre Attack ID are not following the pattern Txxxx: ' + values["name"])
        return v

    @validator('observable')
    def tags_observable(cls,v,values):
        valid_roles = SES_OBSERVABLE_ROLE_MAPPING.keys()
        valid_types = SES_OBSERVABLE_TYPE_MAPPING.keys()
        
        for value in v:
            if value['type'] in valid_types:
                for role in value['role']:
                    if role not in valid_roles:
                        raise ValueError('Observable role ' + role + ' not valid for ' + values["name"] + '. valid options are ' + str(valid_roles))
            else:
                raise ValueError('Observable type ' + value['type'] + ' not valid for ' + values["name"] + '. valid options are ' + str(valid_types))
        return v

    @validator('product')
    def tags_product(cls, v, values):
        valid_products = [
            "Splunk Enterprise", "Splunk Enterprise Security", "Splunk Cloud",
            "Splunk Security Analytics for AWS", "Splunk Behavioral Analytics"
        ]

        for value in v:
            if value not in valid_products:
                raise ValueError('product is not valid for ' + values['name'] + '. valid products are ' + str(valid_products))
        return v

    @validator('risk_score')
    def tags_calculate_risk_score(cls, v, values):
        calculated_risk_score = (int(values['impact']))*(int(values['confidence']))/100
        if calculated_risk_score != int(v):
            raise ValueError('risk_score is calculated wrong: ' + values["name"])
        return v




from contentctl.actions.detection_testing.views.DetectionTestingView import (
    DetectionTestingView,
)

import time
import datetime
import tqdm


class DetectionTestingViewCLI(DetectionTestingView, arbitrary_types_allowed=True):
    pbar: tqdm.tqdm = None
    previous_output_queue_length: int = 0

    def format_pbar(self, completed_detections: int, total_detections: int):
        ratio = f"{completed_detections}/{total_detections}".ljust(9)
        try:
            et = str(self.getRuntime()).ljust(8)
        except Exception as e:
            et = str(e).ljust(8)

        try:
            etr = str(self.getETA()).ljust(8)
        except Exception as e:
            etr = str(e).ljust(8)

        bar = "{percentage:3.2f}%[{bar:30}]"
        self.pbar.bar_format = (
            f"Completed {ratio} {bar} | Elapsed: {et} | Remaining: {etr}"
        )
        self.pbar.reset(total=total_detections)
        self.pbar.update(completed_detections)

    def setup(self):
        self.previous_output_queue_length = len(self.sync_obj.outputQueue)
        self.pbar = tqdm.tqdm(
            total=len(self.sync_obj.inputQueue),
            initial=0,
            bar_format="",
            miniters=0,
            mininterval=0,
        )
        fmt = self.format_pbar(
            len(self.sync_obj.outputQueue), len(self.sync_obj.inputQueue)
        )

        self.showStatus()

    def showStatus(self, interval: int = 1):

        while True:
            summary = self.getSummaryObject()

            total = len(
                summary.get("tested_detections", [])
                + summary.get("untested_detections", [])
                + self.getCurrent()
            )
            self.format_pbar(len(summary.get("tested_detections", [])), total)

            for _ in range(interval):
                if self.sync_obj.terminate:
                    return
                time.sleep(1)

    def showResults(self):
        pass

    def createReport(self):
        pass

    def stop(self):
        self.pbar.close()
        pass





from pydantic import BaseModel, validator, ValidationError

from contentctl.objects.security_content_object import SecurityContentObject



class Macro(BaseModel, SecurityContentObject):
    name: str
    definition: str
    description: str
    arguments: list = None




import uuid
import string

from pydantic import BaseModel, validator, ValidationError

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.playbook_tags import PlaybookTag



class Playbook(BaseModel, SecurityContentObject):
    name: str
    id: str
    version: int
    date: str
    author: str
    type: str
    description: str
    how_to_implement: str
    playbook: str
    check_references: bool = False #Validation is done in order, this field must be defined first
    references: list
    app_list: list
    tags: PlaybookTag





import json
from datetime import datetime
from datetime import timedelta
#import fileinput
import os
import re
import io

class DataManipulation:

    def manipulate_timestamp(self, file_path, sourcetype, source):


        #print('Updating timestamps in attack_data before replaying')
        if sourcetype == 'aws:cloudtrail':
            self.manipulate_timestamp_cloudtrail(file_path)

        if source == 'WinEventLog:System' or source == 'WinEventLog:Security':
            self.manipulate_timestamp_windows_event_log_raw(file_path)

        if source == 'exchange':
            self.manipulate_timestamp_exchange_logs(file_path)


    def manipulate_timestamp_exchange_logs(self, path):
        f = io.open(path, "r", encoding="utf-8")

        first_line = f.readline()
        d = json.loads(first_line)
        latest_event  = datetime.strptime(d["CreationTime"],"%Y-%m-%dT%H:%M:%S")

        now = datetime.now()
        now = now.strftime("%Y-%m-%dT%H:%M:%S")
        now = datetime.strptime(now,"%Y-%m-%dT%H:%M:%S")

        difference = now - latest_event
        f.close()

        #Mimic the behavior of fileinput but in a threadsafe way
        #Rename the file, which fileinput does for inplace.
        #Note that path will now be the new file
        original_backup_file = f"{path}.bak"    
        os.rename(path, original_backup_file)
        
        with open(original_backup_file, "r") as original_file:
            with open(path, "w") as new_file:
                for line in original_file:
                    d = json.loads(line)
                    original_time = datetime.strptime(d["CreationTime"],"%Y-%m-%dT%H:%M:%S")
                    new_time = (difference + original_time)

                    original_time = original_time.strftime("%Y-%m-%dT%H:%M:%S")
                    new_time = new_time.strftime("%Y-%m-%dT%H:%M:%S")
                    #There is no end character appended, no need for end=''
                    new_file.write(line.replace(original_time, new_time))


        os.remove(original_backup_file)

    def manipulate_timestamp_windows_event_log_raw(self, path):

        f = io.open(path, "r", encoding="utf-8")
        self.now = datetime.now()
        self.now = self.now.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        self.now = datetime.strptime(self.now,"%Y-%m-%dT%H:%M:%S.%fZ")

        # read raw logs
        regex = r'\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2} [AP]M'
        data = f.read()
        lst_matches = re.findall(regex, data)
        if len(lst_matches) > 0:
            latest_event  = datetime.strptime(lst_matches[-1],"%m/%d/%Y %I:%M:%S %p")
            self.difference = self.now - latest_event
            f.close()

            result = re.sub(regex, self.replacement_function, data)

            with io.open(path, "w+", encoding='utf8') as f:
                f.write(result)
        else:
            f.close()
            return


    def replacement_function(self, match):
        try:
            event_time = datetime.strptime(match.group(),"%m/%d/%Y %I:%M:%S %p")
            new_time = self.difference + event_time
            return new_time.strftime("%m/%d/%Y %I:%M:%S %p")
        except Exception as e:
            self.logger.error("Error in timestamp replacement occured: " + str(e))
            return match.group()


    def manipulate_timestamp_cloudtrail(self, path):
        

        f = io.open(path, "r", encoding="utf-8")

        try:
            first_line = f.readline()
            d = json.loads(first_line)
            latest_event  = datetime.strptime(d["eventTime"],"%Y-%m-%dT%H:%M:%S.%fZ")

            now = datetime.now()
            now = now.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            now = datetime.strptime(now,"%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            first_line = f.readline()
            d = json.loads(first_line)
            latest_event  = datetime.strptime(d["eventTime"],"%Y-%m-%dT%H:%M:%SZ")

            now = datetime.now()
            now = now.strftime("%Y-%m-%dT%H:%M:%SZ")
            now = datetime.strptime(now,"%Y-%m-%dT%H:%M:%SZ")

        difference = now - latest_event
        f.close()



        #Mimic the behavior of fileinput but in a threadsafe way
        #Rename the file, which fileinput does for inplace.
        #Note that path will now be the new file
        original_backup_file = f"{path}.bak"    
        os.rename(path, original_backup_file)
        
        with open(original_backup_file, "r") as original_file:
            with open(path, "w") as new_file:
                for line in original_file:
                    try:
                        d = json.loads(line)
                        original_time = datetime.strptime(d["eventTime"],"%Y-%m-%dT%H:%M:%S.%fZ")
                        new_time = (difference + original_time)

                        original_time = original_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                        new_time = new_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                        new_file.write(line.replace(original_time, new_time))
                    except ValueError:
                        d = json.loads(line)
                        original_time = datetime.strptime(d["eventTime"],"%Y-%m-%dT%H:%M:%SZ")
                        new_time = (difference + original_time)

                        original_time = original_time.strftime("%Y-%m-%dT%H:%M:%SZ")
                        new_time = new_time.strftime("%Y-%m-%dT%H:%M:%SZ")
                        new_file.write(line.replace(original_time, new_time))


        os.remove(original_backup_file)


from pydantic import BaseModel, validator, ValidationError

from contentctl.objects.security_content_object import SecurityContentObject



class Lookup(BaseModel, SecurityContentObject):
    name: str
    description: str
    collection: str = None
    fields_list: str = None
    filename: str = None
    default_match: str = None
    match_type: str = None
    min_matches: int = None
    case_sensitive_match: str = None



from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingInfrastructure,
)


class DetectionTestingInfrastructureServer(DetectionTestingInfrastructure):
    def start(self):
        pass

    def finish(self):
        super().finish()

    def get_name(self):
        return self.config.container_name



from bottle import route, run, template, Bottle, ServerAdapter
from contentctl.actions.detection_testing.views.DetectionTestingView import (
    DetectionTestingView,
)

from contentctl.objects.unit_test_result import UnitTestResult
from wsgiref.simple_server import make_server, WSGIRequestHandler
import jinja2
import webbrowser
from threading import Thread

DEFAULT_WEB_UI_PORT = 7999

STATUS_TEMPLATE = """
<html>
<head>
<title>contentctl Test {{ percent_complete }}</title>
<script src="https://code.jquery.com/jquery-3.5.1.js"></script>
<script src="https://cdn.datatables.net/1.13.3/js/jquery.dataTables.min.js"></script>
<link rel="stylesheet" href="https://cdn.datatables.net/1.13.3/css/jquery.dataTables.min.css">
<script>
$(document).ready(function () {
    $("#results").DataTable();
    $("#runningTests").DataTable();
});
</script>
</head>
<body>
<table id="runningTests" class="display" style="width:100%">
    <thead>
        <tr>
            <th>Instance Name</th>
            <th>Current Test</th>
            <th>Search</th>
        </tr>
    </thead>
    <tbody>
        {% for containerName, data in currentTestingQueue.items() %}
        <tr>
            <td>{{ containerName }}</td>
            <td>{{ data["name"] }}</td>    
            <td>{{ data["search"] }}</td>    
        </tr>
        {% endfor %}
    </tbody>
</table>

<table id="results" class="display" style="width:100%">
    <thead>
        <tr>
            <th>Test Name</th>
            <th>Test SID</th>
            <th>Run Duration</th>
            <th>Message</th>
            <th>Success</th>
        </tr>
    </thead>
    <tbody>
        {% for detection in detections %}
        {% for test in detection.tests %}
        <tr>
            <td>{{ test.name }}</td>
            <td><a href="{{test.sid_link}}" target="_blank"/>SID</td>
            <td>{{ test.runDuration }}</td>
            <td>{{ test.message }}</td>
            {% if test.success %}
            <td>True</td>
            {% else %}
            <td style="font-weight: bold;background-color: #ff9999"><b>False</b></td>
            {% endif %}
            
        </tr>
        {% endfor %}
        {% endfor %}
    </tbody>
</table>
</body>
</hmtl>
"""


class SimpleWebServer(ServerAdapter):
    server = None

    def run(self, handler):
        class DontLog(WSGIRequestHandler):
            def log_request(*args, **kwargs):
                pass

            def log_exception(*args, **kwargs):
                print(f"Exception in Web View:\n\tARGS:{args}\n\t{kwargs}")

        self.options["handler_class"] = DontLog
        self.server = make_server(
            "localhost", DEFAULT_WEB_UI_PORT, handler, **self.options
        )

        self.server.serve_forever()


class DetectionTestingViewWeb(DetectionTestingView):
    bottleApp: Bottle = Bottle()
    server: SimpleWebServer = SimpleWebServer()

    class Config:
        arbitrary_types_allowed = True

    def setup(self):
        self.bottleApp.route("/", callback=self.showStatus)
        self.bottleApp.route("/status", callback=self.showStatus)
        self.bottleApp.route("/results", callback=self.showResults)
        self.bottleApp.route("/report", callback=self.createReport)

        t = Thread(
            target=self.bottleApp.run, daemon=True, kwargs=({"server": self.server})
        )
        t.start()

        try:
            webbrowser.open(f"http://localhost:{DEFAULT_WEB_UI_PORT}")
        except Exception as e:
            print(f"Could not open webbrowser for status page: {str(e)}")

    def stop(self):

        if self.server.server is None:
            print("Web Server is not running anyway - nothing to shut down")
            return

        # print("called web server shutdown")
        # self.server.server.shutdown()
        # print("finished calling web server shutdown")

    def showStatus(self, interval: int = 60):
        # Status updated on page load
        # get all the finished detections:

        jinja2_template = jinja2.Environment().from_string(STATUS_TEMPLATE)
        summary_dict = self.getSummaryObject(
            test_model_fields=["success", "message", "sid_link"]
        )

        res = jinja2_template.render(
            currentTestingQueue=self.sync_obj.currentTestingQueue,
            percent_complete=summary_dict.get("percent_complete", 0),
            detections=summary_dict["tested_detections"],
        )

        return template(res)

    def showResults(self):
        # Results generated on page load
        return template("page for {{status}}", status="RESULTS")

    def createReport(self):
        # Report generated on page load
        return template("page for {{status}}", status="REPORT")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['contentctl',
 'contentctl.actions',
 'contentctl.actions.detection_testing',
 'contentctl.actions.detection_testing.infrastructures',
 'contentctl.actions.detection_testing.views',
 'contentctl.enrichments',
 'contentctl.helper',
 'contentctl.input',
 'contentctl.objects',
 'contentctl.output']

package_data = \
{'': ['*'],
 'contentctl': ['templates/*',
                'templates/detections/*',
                'templates/macros/*',
                'templates/splunk_app/metadata/*',
                'templates/stories/*',
                'templates/tests/*'],
 'contentctl.output': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'attackcti>=0.3.7,<0.4.0',
 'bottle>=0.12.23,<0.13.0',
 'docker>=6.0.1,<7.0.0',
 'gitpython>=3.1.29,<4.0.0',
 'pycvesearch>=1.2,<2.0',
 'pydantic>=1.10.2,<2.0.0',
 'questionary>=1.10.0,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'semantic-version>=2.10.0,<3.0.0',
 'splunk-sdk>=1.7.2,<2.0.0',
 'tqdm>=4.65.0,<5.0.0',
 'validators>=0.20.0,<0.21.0',
 'xmltodict>=0.13.0,<0.14.0']

entry_points = \
{'console_scripts': ['contentctl = splunk_contentctl.contentctl:main']}

setup_kwargs = {
    'name': 'contentctl',
    'version': '1.0.0',
    'description': 'Splunk Content Control Tool',
    'long_description': '\n# contentctl\n<p align="center">\n<img src="docs/contentctl_logo_white.png" title="In case you\'re wondering, it\'s a capybara" alt="contentctl logo" width="250" height="250"></p>\n\n\n\n\n# Introduction\n#### Security Is Hard \nAnyone who has managed a [SOC](acronym) will tell you it\'s hard work.  SOC Managers, Detection Engineers, and Cybersecurity Professionals must understand and manage countless tools and data sources while ensuring the reliability and security of their network and applications.  At the same time, they must be able to detect and react to data breeches, vulnerabilities, and performance degradation in minutes - not seconds or hours.\nThese responsibilities leave little time for writing new content - let alone documentation, maintenance, and testing of legacy content.  Existing solutions like Wikis, JIRA tickets, Excel Spreadsheets, and "asking that one team member who knows everything" don\'t scale and spread information across various systems.       \n#### contentctl Makes It ~~Easy~~ Less Hard \ncontentctl is a single application that support the full cycle of security content development (each of the links below will redirect to the appropriate section of the README/Wiki):\n\n- [Create Content Packs](contentctl-init) - Version-Controlled collections of searches, documentation, and test data\n- [Add New Content](contentctl-new-content) - searches, macros, lookups, and other content\n- [Statically Validate](contentctl-validate) Content Against a Well-Defined Baseline\n- [Building](contentctl-build) a Content Pack for your target - Use a supported output format or design your own\n- [Generate](contentctl-generate) documentation, MITRE Maps, and a website that makes your Content Pack searchable and easy to understand  \n- [Test](contentctl-test) your Content Pack on a running Splunk Instance\n- [Deploy](contentctl-deploy) your Content Pack to Splunk Cloud Instance, via REST API, or create a Splunkbase Application for manual deployment   \n \n \n \n # Ecosystem\n| Project               | Description                                             |\n| --------------------- | ------------------------------------------------------- |\n| [Splunk Attack Range](https://github.com/splunk/attack_range)          | Easily deploy a preconfigured Splunk Environment locally or on AWS containing a Splunk Instance, Windows and Linux Machines, and Attacker Tools like Kali Linux.  Automatically simulate attacks or run your own|\n| [PurpleSharp Attack Simulation](https://github.com/mvelazc0/PurpleSharp) | Open source adversary simulation tool for Windows Active Directory environments (integrated into Attack Range)|\n| [Red Canary Atomic Red Team](https://github.com/redcanaryco/atomic-red-team)          | Library of attack simulations mapped to the MITRE ATT&CK® framework (integrated into Attack Range)|\n| [Splunk Attack Data](https://github.com/splunk/attack_data)          | Repository of Attack Simulation Data for writing and Testing Detections|                         |\n| [Splunk Security Content](https://github.com/splunk/security_content)          | Splunk Threat Research Team\'s Content included in the [Enterprise Security Content Update App (ESCU)](https://splunkbase.splunk.com/app/3449)|\n| [Splunk contentctl](https://github.com/splunk/contentctl)          | Generate, validate, build, test, and deploy custom Security Content|\n| [SigmaHQ Sigma Rules](https://github.com/SigmaHQ/sigma) | Official Repository for Sigma Rules. These rules are an excellent starting point for new content. |\n| [Other Important Project(s)](https://github.com/otherorg/projectname)          | Additional projects that STRT uses or contributes to|\n\n\n\n\n## Workflow\n```mermaid\ngraph TD\nstarting_state --contentctl init--> contentPack \nnewContent{new_content} --contentctl new sigma content--> contentPack\nnewContent{new_content} --contentctl new spl content--> contentPack\ncontentPack --contentctl validate--> ValidatedContent\nValidatedContent --contentctl build--> SplunkContent\nValidatedContent --contentctl docs--> WebsiteDocumentation\nValidatedContent --contentctl docs--> MITRE_mapDocumentation\nSplunkContent --contentctl test via container-->TestResults\nTestResults --contentctl deploy-->SplunkRESTAPI\nTestResults --contentctl deploy-->SplunkbaseApp\nTestResults --contentctl deploy-->SplunkCloudInstance\n```\n\n## Installation\n### Requirements\nTesting is run using [GitHub Hosted Runners](https://docs.github.com/en/actions/using-github-hosted-runners/about-github-hosted-runners).  Ubuntu22.04 is the recommended configuration.\n| Operating System | Supported | Passing Intregration Tests | \n| ---------------- | --------- | ------ |\n| Windows 10       | Yes       | Not tested yet |\n| Windows 11       | Yes       | Not tested yet |\n| Ununtu 20.04 LTS | Yes       | YES (locally + GitHub Actions) |\n| *Ubuntu 22.04 LTS| Yes       | YES (locally + GitHub Actions) |\n| macOS Montery 12       | Yes       | Not tested yet |\n| macOS Ventura 13       | Yes       | Yes (locally) |\n\n| Requirement | Required | Description |  Passing Integration Tests |\n| --------------------- | ----- | ---- | ------ |\n| Python <3.9 | Not Supported | No support planned.  contentctl tool uses modern language constructs not supported ion Python3.8 and below | N/A |\n| Python 3.9 | Required | contentctl tool is written in Python | Yes (locally + GitHub Actions) |\n| Python 3.10 | Not Supported (yet) | Will be supported | Yes (in GitHub Actions) |\n| Python 3.11 | Not Supported (yet) | Will be supported | Yes (in GitHub Actions)  |\n| Docker (local or remote) | Recommended | A running Splunk Server is required for Dynamic Testing.  contentctl can automatically create, configure, and destroy this server as a Splunk container during the lifetime of a test. | (locally + GitHub Actions) |\n\n\n**Because contentctl uses a large number of dependencies, it\'s recommended to install contentctl in a virtual environment. \nThe instructions below use "poetry" and follow this recommendation.  Poetry will automatically create a new virtual environment for you.  \nIt is typically recommended to install poetry to the Global Python Environment.**\n\n#### Install via pip (coming soon): \n```\npython3.9 -m venv .venv\nsource .venv/bin/activate\npip install splunk-contentctl\n```\n\n#### From Source (advanced)\n```\ngit clone git@github.com:splunk/contentctl.git\ncd contentctl\npython3.9 -m pip install poetry\npoetry install\npoetry shell\ncontentctl --help\n```\n\n\n# Usage\n\n### contentctl init\nCreates a new Content Pack in the current directory as well as a configuration file called [contentctl.yml](contentctl.yml) which contains a number of important configuration options.\nThe content pack contains a wide variety of content types:\n- [detections](detection)\n- [baselines](baseline)\n- [lookups](lookup)\n- [macros](macro)\n- [stories](story)\n\n### contentctl new [--type TYPE]\nChoose TYPE {detection, story} to create new content for the Content Pack.  The tool will interactively ask a series of questions required for generating a basic piece of content and automatically add it to the Content Pack.\n\n### contentctl validate\nPerforms static validation on all of the content in this Content Pack.  Writing validation is extremely complex.  Each piece of content can one or dozens of fields ranging from free text to numbers to references to other content.  contentctl\'s build in validation ensures that a number of conditions are met:\n- Required fields in content are defined\n- Values in fields are appropriate.  For example, if content references a Datamodel, then that Datamodel must be valid. Similarly, numeric values that must fall within a certain range must be validated.\n- If a piece of content references other content, then it exists.  For example, if a detection makes use of a macro then the existence and correctness of the macro must be validated.\n- There are no conflicts between content.  For example, two detections cannot have the same name.\nIf any of these conditions are not met, then a descriptive error will be printed and contentctl will return a nonzero exit code.  This makes it suitable for running in an automated workflow or other CICD context.  Instead of exiting on the first error, _contentctl validate_ will continue validating all content so that it can present all relevant error before exiting.\n\n### contentctl build\nGenerates Content Packs in the output format defined in the [contentctl.yml](contentctl.yml) configuration file.  These outputs may include {splunk_app, api, ba}.  When _contentctl build_ runs, it first performs a _contentctl validate_ in order to ensure that a valid app is generated.  Note that it is *NOT* required or recommended to run _contentctl validate_ separately if the intention is to build a Content Pack.\n\n### contentctl report\n### contentctl inspect\n### contentctl deploy\nThe reason to build content is so that it can be deployed to your environment.  However, deploying content to multiple servers and different types of infrastructure can be tricky and time-consuming.  contentctl makes this easy by supporting a number of different deployment mechanisms. Deployment targets can be defined in [contentctl.yml](contentctl.yml).\n- Deploy via API - Using the REST API, individual pieces of content are deployed to a running server.  This is a great way to deploy all of the content in a content pack, but can also be used to deploy individual peices of content.\n- Deploy to Splunk Cloud via ACS - Using the Automated Private App Vetting (APAV) Feature of the Admin Config Service (ACS), Splunk Cloud customers can easily deploy custom apps to their environments.   \n### contentctl docs\n### contentctl test\nThe static validation performed by *contentctl validate* can only take you so far.  While it\'s powerful, and fast, it can only tell determine if the content is *syntactically* correct.  *contentctl test* can test your content on real Splunk Infrastructure to ensure there are no errors in your SPL, raw data can be properly ingested/processed/accelerated on your server, your search finds the event(s) you\'re looking for in raw data, and even provides high-level runtime performance metrics about your searches.    The following diagram shows this workflow at a high level\n\n```mermaid\ngraph TD\nSplunkContainer--start-->SplunkInstance\nSplunkServer-->SplunkInstance\nSplunkInstance--Install Apps-->ConfiguredInstance\nSplunkInstance--Configure Settings-->ConfiguredInstance\nContentPack--Deploy-->ConfiguredInstance\nConfiguredInstance--Start Test-->Testing\nTesting--Test Detection-->Detection\n\nDetection--Replay Raw Data-->Result\nsubgraph Test Indidivual Detection\n\tDetection--Execute Search-->Result\t\nend\nResult--Cleanup-->Testing\nTesting--Cleanup-->TestResults\n```\n\n#### Testing Modes\nThere are a number of different testing modes that control which content will be tested.  This can be controlled with the `--mode {}` option at the command line\n- all - This will test all of the content in the content pack. Please note that by default only detections marked production will be tested (detections marked as deprecated or experimental, for example, will be ignored).\n- selected - Detections whose relative paths are provided on the command line will be tested.  This is particularly useful if you would like to troubleshoot or update just a handful of detections and can save a significant amount of time.  For example, to test two detections use the following command `contentctl test --mode selected --detections_list detections/detection_one.yml detections/detection_two.yml`\n- **changes - (not currently supported) If you have a large number of detections and use a branching workflow to create new content or update content, then this is an easy way to automatically find and test only that content automatically.  This prevents you from needing to explicitly list the subset of content to test using "selected"** \n\n#### Testing Behavior\ncontentctl test\'s default mode allows it to quickly test all content with requiring user interaction.  This makes it suitable for local tests as well as CI/CD workflows.  However, users often want to troubleshoot a test if it fails.  contentctl allows you to change the tool\'s behavior if and/or when a test fails:\n- --behavior never_pause - The default behavior.  If a test does not pass, the tool begins the next test immediately\n- --behavior pause_on_failure - If a test fails, then additional information about that test, and the raw SPL of the test, is printed to the terminal.  A user may then click (or CMD+Click) the "LINK" to begin interactively debugging the failed test on the Splunk Server.  Note that the credentials for the server are printed out at the very beginning of the test.  After you have finished debugging the failure, hit "Return" in the terminal to move on to the next test. The attack_data for this test remains loaded on the server for debugging until the user moves on to the next test.\n- --behavior always_pause - Similar to pause_on_failure, but this pauses after every test regardless of whether it passes or fails.  \n\n\n1. **init** - Initilialize a new repo from scratch so you can easily add your own content to a custom application. \n2. **new** - Creates new content (detection, story)\n3. **validate** - Validates written content\n4. **build** - Builds an application suitable for deployment on a search head using Slim, the Splunk Packaging Toolkit\n5. **deploy** - Deploy the security content pack to a Splunk Server\n6. **docs** - Create documentation as Markdown\n7. **reporting** - Create different reporting files such as a Mitre ATT&CK overlay\n\n\n\n# Acronyms\n| Acronym | Meaning| Description | \n| --------------------- | ------------------------------------------------------- | ---- |\n| SOC | Security Operation Center | Description of a SoC | \n| DaC | Detection as Code | A systematic approach applying DevOps priciples to Detection Engineering. DaC enables Continuous Integration and Continuous Delivery of Detectionsa via automated validation, testing, and deployment |\n| CICD | Continuous Integration/Continuous Delivery | A modern DevOps practice that encourages users to make small, frequent changes which are automatically tested and deployed. This contrasts with legacy approaches that emphasize large changes which may be manually tested an infrequently deployed. |\n| ACS| Admin Config Service | The [Splunk Admin Configu Service](https://docs.splunk.com/Documentation/SplunkCloud/9.0.2209/Config/ACSIntro) is a cloud-native API that provides programmatic self-service administration capabilities for Splunk Cloud Platform. One of its features, Automated Private App Vetting (APAV) enables the installation of custom app on Splunk Cloud instances. | \n| APAV | Automated Private App Vetting | [Automated Private App Vetting](https://docs.splunk.com/Documentation/SplunkCloud/9.0.2209/Config/ManageApps) enables admins to use the ACS API to install, upgrade, and uninstall apps directly on your Splunk Cloud Platform deployment without assistance from Splunk Support. |\n\n\n# License\nCopyright 2023 Splunk Inc.\n\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at\n\nhttp://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.\n',
    'author': 'STRT',
    'author_email': 'research@splunk.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import json


class JsonWriter():

    @staticmethod
    def writeJsonObject(file_path : str, obj) -> None:

        with open(file_path, 'w') as outfile:
            json.dump(obj, outfile, ensure_ascii=False)


from contentctl.objects.test_config import TestConfig
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingInfrastructure,
)
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructureContainer import (
    DetectionTestingInfrastructureContainer,
)
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructureServer import (
    DetectionTestingInfrastructureServer,
)

from contentctl.objects.app import App
import pathlib
import os
from contentctl.helper.utils import Utils
from urllib.parse import urlparse
import time
from copy import deepcopy
from contentctl.objects.enums import DetectionTestingTargetInfrastructure
import signal
import datetime

# from queue import Queue

CONTAINER_APP_PATH = pathlib.Path("apps")

from dataclasses import dataclass

# import threading
import ctypes
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingInfrastructure,
    DetectionTestingManagerOutputDto,
)
from contentctl.actions.detection_testing.views.DetectionTestingView import (
    DetectionTestingView,
)

from contentctl.objects.enums import PostTestBehavior

from pydantic import BaseModel, Field
from contentctl.input.director import DirectorOutputDto
from contentctl.objects.detection import Detection


import concurrent.futures

import tqdm


@dataclass(frozen=False)
class DetectionTestingManagerInputDto:
    config: TestConfig
    testContent: DirectorOutputDto
    views: list[DetectionTestingView]


class DetectionTestingManager(BaseModel):
    input_dto: DetectionTestingManagerInputDto
    output_dto: DetectionTestingManagerOutputDto
    detectionTestingInfrastructureObjects: list[DetectionTestingInfrastructure] = []

    def setup(self):
        # Some views, such as the Web View, will require some initial setup.
        # for view in self.input_dto.views:
        #    view.setup()

        # for content in self.input_dto.testContent.detections:
        #    self.pending_queue.put(content)
        self.output_dto.inputQueue = self.input_dto.testContent.detections
        self.create_DetectionTestingInfrastructureObjects()

    def execute(self) -> DetectionTestingManagerOutputDto:
        def sigint_handler(signum, frame):
            print("SIGINT (Ctrl-C Received.  Shutting down test...)")
            self.output_dto.terminate = True
            if self.input_dto.config.post_test_behavior in [
                PostTestBehavior.always_pause,
                PostTestBehavior.pause_on_failure,
            ]:
                # It is possible that we are stuck waiting at in input() prompt, so inject
                # a newline '\r\n' which will cause that wait to stop
                print("*******************************")
                print(
                    "If testing is paused and you are debugging a detection, you MUST hit CTRL-D at the prompt to complete shutdown."
                )
                print("*******************************")

        signal.signal(signal.SIGINT, sigint_handler)

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.input_dto.config.num_containers,
        ) as instance_pool, concurrent.futures.ThreadPoolExecutor(
            max_workers=len(self.input_dto.views)
        ) as view_runner, concurrent.futures.ThreadPoolExecutor(
            max_workers=self.input_dto.config.num_containers,
        ) as view_shutdowner:

            # Start all the views
            future_views = {
                view_runner.submit(view.setup): view for view in self.input_dto.views
            }
            # Configure all the instances
            future_instances_setup = {
                instance_pool.submit(instance.setup): instance
                for instance in self.detectionTestingInfrastructureObjects
            }

            # Wait for all instances to be set up
            for future in concurrent.futures.as_completed(future_instances_setup):
                try:
                    result = future.result()
                except Exception as e:
                    self.output_dto.terminate = True
                    print(f"Error setting up container: {str(e)}")

            # Start and wait for all tests to run
            if not self.output_dto.terminate:
                self.output_dto.start_time = datetime.datetime.now()
                future_instances_execute = {
                    instance_pool.submit(instance.execute): instance
                    for instance in self.detectionTestingInfrastructureObjects
                }
                # What for execution to finish
                for future in concurrent.futures.as_completed(future_instances_execute):
                    try:
                        result = future.result()
                    except Exception as e:
                        self.output_dto.terminate = True
                        print(f"Error running in container: {str(e)}")

            self.output_dto.terminate = True

            future_views_shutdowner = {
                view_shutdowner.submit(view.stop): view for view in self.input_dto.views
            }
            for future in concurrent.futures.as_completed(future_views_shutdowner):
                try:
                    result = future.result()
                except Exception as e:
                    print(f"Error stopping view: {str(e)}")

            for future in concurrent.futures.as_completed(future_views):
                try:
                    result = future.result()
                except Exception as e:
                    print(f"Error running container: {str(e)}")

        return self.output_dto

    def create_DetectionTestingInfrastructureObjects(self):
        import sys

        for index in range(self.input_dto.config.num_containers):
            instanceConfig = deepcopy(self.input_dto.config)
            instanceConfig.api_port += index * 2
            instanceConfig.hec_port += index * 2
            instanceConfig.web_ui_port += index

            instanceConfig.container_name = instanceConfig.container_name % (index,)

            if (
                self.input_dto.config.target_infrastructure
                == DetectionTestingTargetInfrastructure.container
            ):

                self.detectionTestingInfrastructureObjects.append(
                    DetectionTestingInfrastructureContainer(
                        config=instanceConfig, sync_obj=self.output_dto
                    )
                )

            elif (
                self.input_dto.config.target_infrastructure
                == DetectionTestingTargetInfrastructure.server
            ):

                self.detectionTestingInfrastructureObjects.append(
                    DetectionTestingInfrastructureServer(
                        config=instanceConfig, sync_obj=self.output_dto
                    )
                )

            else:

                print(
                    f"Unsupported target infrastructure '{self.input_dto.config.target_infrastructure}'"
                )
                sys.exit(1)



import enum


class AnalyticsType(enum.Enum):
    TTP = 1
    anomaly = 2
    hunting = 3
    correlation = 4


class DataModel(enum.Enum):
    Endpoint = 1
    Network_Traffic = 2
    Authentication = 3
    Change = 4
    Change_Analysis = 5
    Email = 6
    Network_Resolution = 7
    Network_Sessions = 8
    UEBA = 9
    Updates = 10
    Vulnerabilities = 11
    Web = 12
    Endpoint_Processes = 13
    Endpoint_Filesystem = 14
    Endpoint_Registry = 15
    Risk = 16
    Splunk_Audit = 17


class SecurityContentType(enum.Enum):
    detections = 1
    baselines = 2
    stories = 3
    playbooks = 4
    macros = 5
    lookups = 6
    deployments = 7
    investigations = 8
    unit_tests = 9


class SecurityContentProduct(enum.Enum):
    # This covers ESCU as well as other apps initialized
    # by splunk_security_content_builder
    splunk_app = "splunk_app"
    ba_objects = "ba_objects"
    json_objects = "json_objects"


class LogLevel(enum.Enum):
    NONE = "NONE"
    ERROR = "ERROR"
    INFO = "INFO"


class AlertActions(enum.Enum):
    notable = "notable"
    rba = "rba"
    email = "email"


class PostTestBehavior(str, enum.Enum):
    always_pause = "always_pause"
    pause_on_failure = "pause_on_failure"
    never_pause = "never_pause"


class DetectionTestingMode(str, enum.Enum):
    selected = "selected"
    all = "all"
    changes = "changes"


class DetectionTestingTargetInfrastructure(str, enum.Enum):
    container = "container"
    server = "server"


class InstanceState(str, enum.Enum):
    starting = "starting"
    running = "running"
    error = "error"
    stopping = "stopping"
    stopped = "stopped"



from pydantic import BaseModel
import abc
from typing import Callable
from contentctl.objects.test_config import TestConfig
import datetime
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingManagerOutputDto,
)
from contentctl.helper.utils import Utils


class DetectionTestingView(BaseModel, abc.ABC):
    config: TestConfig
    sync_obj: DetectionTestingManagerOutputDto

    interval: float = 10
    next_update: float = 0

    def setup(self):
        pass

    def stop(self):
        pass

    def showStatus(self, interval: int = 60):
        pass

    def showResults(self):
        pass

    def createReport(self):
        pass

    def getCurrent(self) -> list[str]:
        return [
            d
            for d in self.sync_obj.currentTestingQueue
            if self.sync_obj.currentTestingQueue[d] is not None
        ]

    def getRuntime(self) -> datetime.timedelta:
        if self.sync_obj.start_time == None:
            raise Exception("Unknown Time")
        runtime = datetime.datetime.now() - self.sync_obj.start_time
        runtime -= datetime.timedelta(microseconds=runtime.microseconds)
        return runtime

    def getETA(self) -> datetime.timedelta:
        summary = self.getSummaryObject()

        num_tested = len(summary.get("tested_detections", []))
        num_untested = len(summary.get("untested_detections", []))

        if num_tested == 0:
            raise Exception("Unknown ETA")
        elif num_untested == 0:
            raise Exception("Finishing test")

        try:
            runtime = self.getRuntime()
            time_per_detection = runtime / num_tested
            remaining_time = num_untested * time_per_detection
            remaining_time -= datetime.timedelta(
                microseconds=remaining_time.microseconds
            )
        except:
            raise Exception("Unknown ETA")
        return remaining_time

    def getSummaryObject(
        self,
        test_model_fields: list[str] = ["success", "message"],
        test_job_fields: list[str] = ["resultCount", "runDuration"],
    ) -> dict:
        total_untested = len(self.sync_obj.inputQueue)

        tested_detections = []
        total_pass = 0
        total_fail = 0
        for detection in self.sync_obj.outputQueue:
            summary = detection.get_summary(
                test_job_fields=test_job_fields, test_model_fields=test_model_fields
            )
            if summary["success"] == True:
                total_pass += 1
            else:
                total_fail += 1
            tested_detections.append(summary)
        # All failures appear first
        tested_detections.sort(key=lambda x: (x["success"], x["name"]))

        untested_detections = []
        for detection in self.sync_obj.inputQueue:
            untested_detections.append(detection.get_summary())
        # All failures appear first
        untested_detections.sort(key=lambda x: x["name"])

        if (total_fail + len(untested_detections)) == 0:
            overall_success = True
        else:
            overall_success = False

        percent_complete = Utils.getPercent(
            len(tested_detections), len(untested_detections), 1
        )
        success_rate = Utils.getPercent(
            total_pass, total_fail + total_pass + total_untested, 1
        )

        result_dict = {
            "summary": {
                "success": overall_success,
                "total_detections": total_pass + total_fail,
                "total_pass": total_pass,
                "total_fail_or_untested": total_fail + total_untested,
                "success_rate": success_rate,
            },
            "tested_detections": tested_detections,
            "untested_detections": untested_detections,
            "percent_complete": percent_complete,
        }
        return result_dict



import os
import asyncio
import sys

from pathlib import Path

from contentctl.objects.enums import SecurityContentType
from contentctl.output.jinja_writer import JinjaWriter


class DocMdOutput():
    index = 0
    files_to_write = 0
    
    def writeObjects(self, objects: list, output_path: str) -> None:
        self.files_to_write = sum([len(obj) for obj in objects])
        self.index = 0
        progress_percent = ((self.index+1)/self.files_to_write) * 100
        if (sys.stdout.isatty() and sys.stdin.isatty() and sys.stderr.isatty()):
            print(f"\r{'Docgen Progress'.rjust(23)}: [{progress_percent:3.0f}%]...", end="", flush=True)

        attack_tactics = set()
        datamodels = set()
        categories = set()
        for story in objects[0]:
            if story.tags.category:
                categories.update(story.tags.category)

        for detection in objects[1]:
            if detection.tags.mitre_attack_enrichments:
                for attack in detection.tags.mitre_attack_enrichments:
                    attack_tactics.update(attack.mitre_attack_tactics)

            if detection.datamodel:
                datamodels.update(detection.datamodel)
        
        Path(os.path.join(output_path, 'overview')).mkdir(parents=True, exist_ok=True)
        Path(os.path.join(output_path, 'detections')).mkdir(parents=True, exist_ok=True)
        Path(os.path.join(output_path, 'stories')).mkdir(parents=True, exist_ok=True)
        Path(os.path.join(output_path, 'playbooks')).mkdir(parents=True, exist_ok=True)

        JinjaWriter.writeObjectsList('doc_story_page.j2', os.path.join(output_path, 'overview/stories.md'), sorted(objects[0], key=lambda x: x.name))
        self.writeObjectsMd(objects[0], os.path.join(output_path, 'stories'), 'doc_stories.j2')

        JinjaWriter.writeObjectsList('doc_detection_page.j2', os.path.join(output_path, 'overview/detections.md'), sorted(objects[1], key=lambda x: x.name))
        self.writeObjectsMd(objects[1], os.path.join(output_path, 'detections'), 'doc_detections.j2')

        JinjaWriter.writeObjectsList('doc_playbooks_page.j2', os.path.join(output_path, 'overview/paybooks.md'), sorted(objects[2], key=lambda x: x.name))
        self.writeObjectsMd(objects[2], os.path.join(output_path, 'playbooks'), 'doc_playbooks.j2')
        
        print("Done!")
    
    
    # def writeNavigationPageObjects(self, objects: list, output_path: str) -> None:
    #     for obj in objects:
    #         JinjaWriter.writeObject('doc_navigation_pages.j2', os.path.join(output_path, '_pages', obj.lower().replace(' ', '_') + '.md'),
    #             {
    #                 'name': obj
    #             }
    #         )

    def writeObjectsMd(self, objects, output_path: str, template_name: str) -> None:
        for obj in objects:
            progress_percent = ((self.index+1)/self.files_to_write) * 100
            self.index+=1
            if (sys.stdout.isatty() and sys.stdin.isatty() and sys.stderr.isatty()):
                print(f"\r{'Docgen Progress'.rjust(23)}: [{progress_percent:3.0f}%]...", end="", flush=True)

            JinjaWriter.writeObject(template_name, os.path.join(output_path, obj.name.lower().replace(' ', '_') + '.md'), obj)





import csv
import os
from posixpath import split
from typing import Optional
import sys
from attackcti import attack_client
from functools import cache
import logging
logging.getLogger('taxii2client').setLevel(logging.CRITICAL)


class AttackEnrichment():

    @classmethod

    def get_attack_lookup(self, input_path: str) -> dict:
        print("Getting MITRE Attack Enrichment Data. This may take some time...")
        attack_lookup = dict()
        file_path = os.path.join(input_path, "lookups", "mitre_enrichment.csv")

        try:

            print(f"\r{'Client'.rjust(23)}: [{0:3.0f}%]...", end="", flush=True)
            lift = attack_client()
            print(f"\r{'Client'.rjust(23)}: [{100:3.0f}%]...Done!", end="\n", flush=True)
            
            print(f"\r{'Enterprise'.rjust(23)}: [{0.0:3.0f}%]...", end="", flush=True)
            all_enterprise = lift.get_enterprise(stix_format=False)
            print(f"\r{'Enterprise'.rjust(23)}: [{100:3.0f}%]...Done!", end="\n", flush=True)
            
            print(f"\r{'Relationships'.rjust(23)}: [{0.0:3.0f}%]...", end="", flush=True)
            enterprise_relationships = lift.get_enterprise_relationships()
            print(f"\r{'Relationships'.rjust(23)}: [{100:3.0f}%]...Done!", end="\n", flush=True)
            
            print(f"\r{'Groups'.rjust(23)}: [{0:3.0f}%]...", end="", flush=True)
            enterprise_groups = lift.get_enterprise_groups()
            print(f"\r{'Groups'.rjust(23)}: [{100:3.0f}%]...Done!", end="\n", flush=True)
            
            for index, technique in enumerate(all_enterprise['techniques']):
                progress_percent = ((index+1)/len(all_enterprise['techniques'])) * 100
                if (sys.stdout.isatty() and sys.stdin.isatty() and sys.stderr.isatty()):
                    print(f"\r\t{'MITRE Technique Progress'.rjust(23)}: [{progress_percent:3.0f}%]...", end="", flush=True)
                apt_groups = []
                for relationship in enterprise_relationships:
                    if (relationship['target_ref'] == technique['id']) and relationship['source_ref'].startswith('intrusion-set'):
                        for group in enterprise_groups:
                            if relationship['source_ref'] == group['id']:
                                apt_groups.append(group['name'])

                tactics = []
                if ('tactic' in technique):
                    for tactic in technique['tactic']:
                        tactics.append(tactic.replace('-',' ').title())

                if not ('revoked' in technique):
                    attack_lookup[technique['technique_id']] = {'technique': technique['technique'], 'tactics': tactics, 'groups': apt_groups}

            f = open(file_path, 'w')
            writer = csv.writer(f)
            writer.writerow(['mitre_id', 'technique', 'tactics' ,'groups'])
            for key in attack_lookup.keys():
                if len(attack_lookup[key]['groups']) == 0:
                    groups = 'no'
                else:
                    groups = '|'.join(attack_lookup[key]['groups'])
                
                writer.writerow([
                    key,
                    attack_lookup[key]['technique'],
                    '|'.join(attack_lookup[key]['tactics']),
                    groups
                ])
            
            f.close()

        except Exception as err:
            print('Warning: ' + str(err))
            print('Use local copy lookups/mitre_enrichment.csv')
            dict_from_csv = {}
            with open(file_path, mode='r') as inp:
                reader = csv.reader(inp)
                attack_lookup = {rows[0]:{'technique': rows[1], 'tactics': rows[2].split('|'), 'groups': rows[3].split('|')} for rows in reader}
            attack_lookup.pop('mitre_id')

        print("Done!")
        return attack_lookup


import os
import uuid
import questionary
from dataclasses import dataclass
from datetime import datetime

from contentctl.objects.enums import SecurityContentType
from contentctl.input.new_content_questions import NewContentQuestions


@dataclass(frozen=True)
class NewContentGeneratorInputDto:
    type: SecurityContentType
    

@dataclass(frozen=True)
class NewContentGeneratorOutputDto:
    obj: dict


class NewContentGenerator():

    
    def __init__(self, output_dto: NewContentGeneratorOutputDto) -> None:
        self.output_dto = output_dto


    def execute(self, input_dto: NewContentGeneratorInputDto) -> None:
        if input_dto.type == SecurityContentType.detections:
            questions = NewContentQuestions.get_questions_detection()
            answers = questionary.prompt(questions)
            self.output_dto.obj['name'] = answers['detection_name']
            self.output_dto.obj['id'] = str(uuid.uuid4())
            self.output_dto.obj['version'] = 1
            self.output_dto.obj['date'] = datetime.today().strftime('%Y-%m-%d')
            self.output_dto.obj['author'] = answers['detection_author']
            self.output_dto.obj['type'] = answers['detection_type']
            self.output_dto.obj['datamodel'] = answers['datamodels']
            self.output_dto.obj['datamodel'] = answers['datamodels']
            self.output_dto.obj['description'] = 'UPDATE_DESCRIPTION'   
            file_name = self.output_dto.obj['name'].replace(' ', '_').replace('-','_').replace('.','_').replace('/','_').lower()
            self.output_dto.obj['search'] = answers['detection_search'] + ' | `' + file_name + '_filter`'
            self.output_dto.obj['how_to_implement'] = 'UPDATE_HOW_TO_IMPLEMENT'
            self.output_dto.obj['known_false_positives'] = 'UPDATE_KNOWN_FALSE_POSITIVES'            
            self.output_dto.obj['references'] = ['REFERENCE']
            self.output_dto.obj['tags'] = dict()
            self.output_dto.obj['tags']['analytic_story'] = ['UPDATE_STORY_NAME']
            self.output_dto.obj['tags']['asset_type'] = 'UPDATE asset_type'
            self.output_dto.obj['tags']['cis20'] = ['CIS 3', 'CIS 5', 'CIS 16']
            self.output_dto.obj['tags']['confidence'] = 'UPDATE value between 1-100'
            self.output_dto.obj['tags']['context'] = ['Update context']
            self.output_dto.obj['tags']['dataset'] = ['UPDATE_DATASET_URL']
            self.output_dto.obj['tags']['impact'] = 'UPDATE value between 1-100'
            self.output_dto.obj['tags']['kill_chain_phases'] = answers['kill_chain_phases']
            self.output_dto.obj['tags']['message'] = 'UPDATE message'
            self.output_dto.obj['tags']['mitre_attack_id'] = [x.strip() for x in answers['mitre_attack_ids'].split(',')]
            self.output_dto.obj['tags']['nist'] = ['DE.CM']
            self.output_dto.obj['tags']['observable'] = [{'name': 'UPDATE', 'type': 'UPDATE', 'role': ['UPDATE']}]
            self.output_dto.obj['tags']['product'] = ['Splunk Enterprise','Splunk Enterprise Security','Splunk Cloud']
            self.output_dto.obj['tags']['required_fields'] = ['UPDATE']
            self.output_dto.obj['tags']['risk_score'] = 'UPDATE (impact * confidence)/100'
            self.output_dto.obj['tags']['security_domain'] = answers['security_domain']
            #self.output_dto.obj['source'] = answers['detection_kind']
        

        elif input_dto.type == SecurityContentType.stories:
            questions = NewContentQuestions.get_questions_story()
            answers = questionary.prompt(questions)
            self.output_dto.obj['name'] = answers['story_name']
            self.output_dto.obj['id'] = str(uuid.uuid4())
            self.output_dto.obj['version'] = 1
            self.output_dto.obj['date'] = datetime.today().strftime('%Y-%m-%d')
            self.output_dto.obj['author'] = answers['story_author']
            self.output_dto.obj['description'] = 'UPDATE_DESCRIPTION'
            self.output_dto.obj['narrative'] = 'UPDATE_NARRATIVE'
            self.output_dto.obj['references'] = []
            self.output_dto.obj['tags'] = dict()
            self.output_dto.obj['tags']['analytic_story'] = self.output_dto.obj['name']
            self.output_dto.obj['tags']['category'] = answers['category']
            self.output_dto.obj['tags']['product'] = ['Splunk Enterprise','Splunk Enterprise Security','Splunk Cloud']
            self.output_dto.obj['tags']['usecase'] = answers['usecase']


# Needed for a staticmethod to be able to return an instance of the class it belongs to
from __future__ import annotations


import pathlib
import re
import os

from pydantic import BaseModel, validator, ValidationError, Extra, Field
from dataclasses import dataclass
from datetime import datetime
from typing import Union
import validators
from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.enums import DataModel
from contentctl.helper.utils import Utils
import yaml

SPLUNKBASE_URL = "https://splunkbase.splunk.com/app/{uid}/release/{release}/download"


class App(BaseModel, extra=Extra.forbid):

    # uid is a numeric identifier assigned by splunkbase, so
    # homemade applications will not have this
    uid: Union[int, None]

    # appid is basically the internal name of you app
    appid: str

    # Title is the human readable name for your application
    title: str

    # Self explanatory
    description: Union[str, None]
    release: str

    local_path: Union[str, None]
    http_path: Union[str, None]
    # Splunkbase path is made of the combination of uid and release fields
    splunkbase_path: Union[str, None]

    # Ultimate source of the app. Can be a local path or a Splunkbase Path.
    # This will be set via a function call and should not be provided in the YML
    # Note that this is the path relative to the container mount
    environment_path: str = "ENVIRONMENT_PATH_NOT_SET"

    def configure_app_source_for_container(
        self,
        splunkbase_username: Union[str, None],
        splunkbase_password: Union[str, None],
        apps_directory: pathlib.Path,
        container_mount_path: pathlib.Path,
    ):

        splunkbase_creds_provided = (
            splunkbase_username is not None and splunkbase_password is not None
        )

        if splunkbase_creds_provided and self.splunkbase_path is not None:
            self.environment_path = self.splunkbase_path

        elif self.local_path is not None:
            # local path existence already validated
            filename = pathlib.Path(self.local_path)
            destination = str(apps_directory / filename.name)
            Utils.copy_local_file(self.local_path, destination, verbose_print=True)
            self.environment_path = str(container_mount_path / filename.name)

        elif self.http_path is not None:
            from urllib.parse import urlparse

            path_on_server = str(urlparse(self.http_path).path)
            filename = pathlib.Path(path_on_server)
            download_path = str(apps_directory / filename.name)
            Utils.download_file_from_http(self.http_path, download_path)
            self.environment_path = str(container_mount_path / filename.name)

        else:
            raise (
                Exception(
                    f"Unable to download app {self.title}:\n"
                    f"Splunkbase Path : {self.splunkbase_path}\n"
                    f"local_path      : {self.local_path}\n"
                    f"http_path       : {self.http_path}\n"
                    f"Splunkbase Creds: {splunkbase_creds_provided}\n"
                )
            )

    @staticmethod
    def validate_string_alphanumeric_with_underscores(input: str) -> bool:
        if len(input) == 0:
            raise (ValueError(f"String was length 0"))

        for letter in input:
            if not (letter.isalnum() or letter in "_-"):
                raise (
                    ValueError(
                        f"String '{input}' can only contain alphanumeric characters, underscores, and hyphens."
                    )
                )
        return True

    @validator("uid", always=True)
    def validate_uid(cls, v):
        return v

    @validator("appid", always=True)
    def validate_appid(cls, v):
        # Called function raises exception on failure, so we don't need to raise it here
        cls.validate_string_alphanumeric_with_underscores(v)
        return v

    @validator("title", always=True)
    def validate_title(cls, v):
        # Basically, a title can be any string
        return v

    @validator("description", always=True)
    def validate_description(cls, v):
        # description can be anything
        return v

    @validator("release", always=True)
    def validate_release(cls, v):
        # release can be any string
        return v

    @validator("local_path", always=True)
    def validate_local_path(cls, v):
        if v is not None:
            p = pathlib.Path(v)
            if not p.exists():
                raise (ValueError(f"The path local_path {p} does not exist"))
            elif not p.is_file():
                raise (ValueError(f"The path local_path {p} exists, but is not a file"))

        # release can be any string
        return v

    @validator("http_path", always=True)
    def validate_http_path(cls, v, values):
        if v is not None:
            try:
                if bool(validators.url(v)) == False:
                    raise ValueError(f"URL '{v}' is not a valid URL")
            except Exception as e:
                raise (ValueError(f"Error validating the http_path: {str(e)}"))
        return v

    @validator("splunkbase_path", always=True)
    def validate_splunkbase_path(cls, v, values):
        Utils.check_required_fields(
            "splunkbase_path", values, ["local_path", "http_path", "uid", "title"]
        )

        if v is not None:
            try:
                res = bool(validator.url(v))
                if res is False:
                    raise Exception
            except Exception as e:
                raise (ValueError(f"splunkbase_url {v} is not a valid URL"))

            if (
                bool(
                    re.match(
                        "^https://splunkbase\.splunk\.com/app/\d+/release/.+/download$",
                        v,
                    )
                )
                == False
            ):
                raise (
                    ValueError(
                        f"splunkbase_url {v} does not match the format {SPLUNKBASE_URL}"
                    )
                )

        # Try to form the URL and error out if Splunkbase is the only place to get the app
        if values["uid"] is None:
            if values["must_download_from_splunkbase"]:
                raise (
                    ValueError(
                        f"Error building splunkbase_url. Attempting to"
                        f" build the url for '{values['title']}', but no "
                        f"uid was supplied."
                    )
                )
            else:
                return None

        if values["release"] is None:
            if values["must_download_from_splunkbase"]:
                raise (
                    ValueError(
                        f"Error building splunkbase_url. Attempting to"
                        f" build the url for '{values['title']}', but no "
                        f"release was supplied."
                    )
                )
            else:
                return None
        return SPLUNKBASE_URL.format(uid=values["uid"], release=values["release"])

    @staticmethod
    def get_default_apps() -> list[App]:
        all_app_objs: list[App] = []
        with open(
            os.path.join(os.path.dirname(__file__), "../", "templates/app_default.yml"),
            "r",
        ) as app_data:
            all_apps_raw = yaml.safe_load(app_data)
            for a in all_apps_raw:
                app_obj = App.parse_obj(a)
                all_app_objs.append(app_obj)
        return all_app_objs



from pydantic import BaseModel, validator, ValidationError
from contentctl.helper.utils import Utils
from typing import Union


class UnitTestAttackData(BaseModel):
    file_name: str
    data: str
    source: str
    sourcetype: str
    update_timestamp: bool = False
    custom_index: Union[str, None] = None
    host: Union[str, None] = None

    @validator("data", always=True)
    def validate_data(cls, v, values):
        return v
        try:
            Utils.verify_file_exists(v)
        except Exception as e:
            raise (ValueError(f"Cannot find file {v}: {str(e)}"))
        return v



import csv
import glob
import logging
import os
import pathlib
import subprocess
import sys
from typing import Union, Tuple
from docker import types
import datetime
import git
import yaml
from git.objects import base

from contentctl.objects.detection import Detection
from contentctl.objects.story import Story
from contentctl.objects.baseline import Baseline
from contentctl.objects.investigation import Investigation
from contentctl.objects.playbook import Playbook
from contentctl.objects.macro import Macro
from contentctl.objects.lookup import Lookup
from contentctl.objects.unit_test import UnitTest

from contentctl.objects.enums import DetectionTestingMode
import random
import pathlib
from contentctl.helper.utils import Utils

from contentctl.objects.test_config import TestConfig
from contentctl.actions.generate import DirectorOutputDto

# Logger
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
LOGGER = logging.getLogger(__name__)


SSA_PREFIX = "ssa___"


class GithubService:
    def get_all_content(self, director: DirectorOutputDto) -> DirectorOutputDto:
        # get a new director that will be used for testing.
        return DirectorOutputDto(
            self.get_detections(director),
            self.get_stories(director),
            self.get_baselines(director),
            self.get_investigations(director),
            self.get_playbooks(director),
            self.get_macros(director),
            self.get_lookups(director),
            [],
        )

    def get_stories(self, director: DirectorOutputDto) -> list[Story]:
        stories: list[Story] = []
        return stories

    def get_baselines(self, director: DirectorOutputDto) -> list[Baseline]:
        baselines: list[Baseline] = []
        return baselines

    def get_investigations(self, director: DirectorOutputDto) -> list[Investigation]:
        investigations: list[Investigation] = []
        return investigations

    def get_playbooks(self, director: DirectorOutputDto) -> list[Playbook]:
        playbooks: list[Playbook] = []
        return playbooks

    def get_macros(self, director: DirectorOutputDto) -> list[Macro]:
        macros: list[Macro] = []
        return macros

    def get_lookups(self, director: DirectorOutputDto) -> list[Lookup]:
        lookups: list[Lookup] = []
        return lookups

    def get_detections(self, director: DirectorOutputDto) -> list[Detection]:
        if self.config.mode == DetectionTestingMode.selected:
            return self.get_detections_selected(director)
        elif self.config.mode == DetectionTestingMode.all:
            return self.get_detections_all(director)
        elif self.config.mode == DetectionTestingMode.changes:
            return self.get_detections_changed(director)
        else:
            raise (
                Exception(
                    f"Error: Unsupported detection testing mode in GithubServer: {self.config.mode}"
                )
            )

    def get_detections_selected(self, director: DirectorOutputDto) -> list[Detection]:
        detections_to_test: list[Detection] = []
        requested_set = set(self.requested_detections)
        missing_detections: set[pathlib.Path] = set()

        for requested in requested_set:
            matching = list(
                filter(
                    lambda detection: pathlib.Path(detection.file_path).resolve()
                    == requested.resolve(),
                    director.detections,
                )
            )
            if len(matching) == 1:
                detections_to_test.append(matching.pop())
            elif len(matching) == 0:
                missing_detections.add(requested)
            else:
                raise (
                    Exception(
                        f"Error: multiple detection files found when attemping to resolve [{str(requested)}]"
                    )
                )

        if len(missing_detections) > 0:
            missing_detections_str = "\n\t - ".join(
                [str(path.absolute()) for path in missing_detections]
            )
            print(director.detections)
            raise (
                Exception(
                    f"Failed to find the following detection file(s) for testing:\n\t - {missing_detections_str}"
                )
            )

        return detections_to_test

    def get_detections_all(self, director: DirectorOutputDto) -> list[Detection]:
        # Assume we don't need to remove anything, like deprecated or experimental from this
        return director.detections

    def get_detections_changed(self, director: DirectorOutputDto) -> list[Detection]:
        if self.repo is None:
            raise (
                Exception(
                    f"Error: self.repo must be initialized before getting changed detections."
                )
            )
        raise (Exception("not implemented"))
        return []

    def __init__(self, config: TestConfig):
        self.repo = None
        self.requested_detections: list[pathlib.Path] = []
        self.config = config

        if config.mode == DetectionTestingMode.selected:
            if config.detections_list is None or len(config.detections_list) < 1:
                raise (
                    Exception(
                        f"Error: detection mode [{config.mode}] REQUIRES that [{config.detections_list}] contains 1 or more detections, but the value is [{config.detections_list}]"
                    )
                )
            else:
                # Ensure that all of the detections exist
                missing_files = [
                    detection
                    for detection in config.detections_list
                    if not pathlib.Path(detection).is_file()
                ]
                if len(missing_files) > 0:
                    missing_string = "\n\t - ".join(missing_files)
                    raise (
                        Exception(
                            f"Error: The following detection(s) test do not exist:\n\t - {missing_files}"
                        )
                    )
                else:
                    self.requested_detections = [
                        pathlib.Path(detection_file_name)
                        for detection_file_name in config.detections_list
                    ]
                    return

        elif config.mode == DetectionTestingMode.changes:
            # Changes is ONLY possible if the app is version controlled
            # in a github repo.  Ensure that this is the case and, if not
            # raise an exception
            raise (Exception("Mode [changes] is not yet supported."))
            try:
                repo = git.Repo(config.repo_path)
            except Exception as e:
                raise (
                    Exception(
                        f"Error: detection mode [{config.mode}] REQUIRES that [{config.repo_path}] is a git repository, but it is not."
                    )
                )
            if config.main_branch == config.test_branch:
                raise (
                    Exception(
                        f"Error: test_branch [{config.test_branch}] is the same as the main_branch [{config.main_branch}]. When using mode [{config.mode}], these two branches MUST be different."
                    )
                )

            # Ensure that the test branch is checked out
            if self.repo.active_branch.name != config.test_branch:
                raise (
                    Exception(
                        f"Error: detection mode [{config.mode}] REQUIRES that the test_branch [{config.test_branch}] be checked out at the beginning of the test, but it is not."
                    )
                )

            # Ensure that the base branch exists

            if Utils.validate_git_branch_name(
                config.repo_path, "NO_URL", config.main_branch
            ):
                return

        elif config.mode == DetectionTestingMode.all:
            return
        else:
            raise (
                Exception(
                    f"Unsupported detection testing mode [{config.mode}].  Supported detection testing modes are [{DetectionTestingMode._member_names_}]"
                )
            )

    def __init2__(self, config: TestConfig):

        self.repo = git.Repo(config.repo_path)

        if self.repo.active_branch.name != config.test_branch:
            print(
                f"Error - test_branch is '{config.test_branch}', but the current active branch in '{config.repo_path}' is '{self.repo.active_branch}'. Checking out the branch you specified..."
            )
            self.repo.git.checkout(config.test_branch)

        self.config = config

    def clone_project(self, url, project, branch):
        LOGGER.info(f"Clone Security Content Project")
        repo_obj = git.Repo.clone_from(url, project, branch=branch)
        return repo_obj

    def get_detections_to_test(
        self,
        config: TestConfig,
        director: DirectorOutputDto,
        ignore_experimental: bool = True,
        ignore_deprecated: bool = True,
        ignore_ssa: bool = True,
        allowed_types: list[str] = ["Anomaly", "Hunting", "TTP"],
    ) -> list[Detection]:

        print(f"Total detections found: {len(director.detections)}")

        if ignore_experimental:
            director.detections = [
                d for d in director.detections if not (d.experimental == True)
            ]
        if ignore_deprecated:
            director.detections = [
                d for d in director.detections if not (d.deprecated == True)
            ]
        if ignore_ssa:
            director.detections = [
                d
                for d in director.detections
                if not pathlib.Path(d.file_path).name.startswith(SSA_PREFIX)
            ]

        print(
            f"Total detections loaded after removal of experimental, deprecated, and ssa: {len(director.detections)}"
        )

        # Downselect to only the types we want to test. For example, this will by default remove the Correlation type
        director.detections = [
            d for d in director.detections if d.type in allowed_types
        ]

        if config.mode == DetectionTestingMode.changes:

            untracked_files, changed_files = self.get_all_modified_content(
                director.detections
            )
            newline_tab = "\n\t"
            print(
                f"Found the following untracked files:\n\t{newline_tab.join([f.file_path for f in untracked_files])}"
            )
            print(
                f"Found the following modified  files:\n\t{newline_tab.join([f.file_path for f in changed_files])}"
            )

            director.detections = untracked_files + changed_files

        elif config.mode == DetectionTestingMode.all:
            # Don't need to do anything, we don't need to remove it from the list
            pass
        elif config.mode == DetectionTestingMode.selected:
            if config.detections_list is None:
                # We should never get here because validation should catch it.  Adding this test to avoid
                # type warning
                raise (
                    Exception(
                        f"Detection Testing mode is {config.mode}. but Detections List was {config.detections_list}"
                    )
                )

            selected_set = set(
                os.path.join(config.repo_path, d) for d in config.detections_list
            )
            all_detections_set = set([d.file_path for d in director.detections])
            difference = selected_set - all_detections_set
            if len(difference) > 0:
                newline = "\n * "
                print(list(all_detections_set)[:10])
                raise (
                    Exception(
                        f"The detections in the detections_list do not exist:{newline}{newline.join(difference)}"
                    )
                )

            # All the detections exist, so find them an update the objects to reflect them
            director.detections = [
                d for d in director.detections if d.file_path in selected_set
            ]
        else:
            raise (
                Exception(
                    f"Unsupported mode {config.mode}.  Supported modes are {DetectionTestingMode._member_names_}"
                )
            )

        print(f"Finally the number is: {len(director.detections)}")

        if config.mode != DetectionTestingMode.selected:
            # If the user has selected specific detections to run, then
            # run those in that specific order.  Otherwise, shuffle the order.
            # This is particulary important when doing a mock because, for example,
            # we don't want one container to get a group of cloud detections which may,
            # on average, run for longer than the group of endpoint detections on
            # another container
            random.shuffle(director.detections)

        return director.detections

    def get_all_modified_content(
        self,
        detections: list[Detection],
        paths: list[pathlib.Path] = [
            pathlib.Path("detections/"),
            pathlib.Path("tests/"),
        ],
    ) -> Tuple[list[Detection], list[Detection]]:
        # Note that at present, we only search in the 'detections' and 'tests' folders.  In the future, we could search in all
        # folders, for example to evaluate any content affected by a macro or playbook change.

        try:

            # Because we have not passed -all as a kwarg, we will have a MAX of one commit returned:
            # https://gitpython.readthedocs.io/en/stable/reference.html?highlight=merge_base#git.repo.base.Repo.merge_base
            base_commits = self.repo.merge_base(
                self.config.main_branch, self.config.test_branch
            )
            if len(base_commits) == 0:
                raise (
                    Exception(
                        f"Error, main branch '{self.config.main_branch}' and test branch '{self.config.test_branch}' do not share a common ancestor"
                    )
                )
            base_commit = base_commits[0]
            if base_commit is None:
                raise (
                    Exception(
                        f"Error, main branch '{self.config.main_branch}' and test branch '{self.config.test_branch}' common ancestor commit was 'None'"
                    )
                )

            all_changes = base_commit.diff(
                self.config.test_branch, paths=[str(path) for path in paths]
            )

            # distill changed files down to the paths of added or modified files
            all_changes_paths = [
                os.path.join(self.config.repo_path, change.b_path)
                for change in all_changes
                if change.change_type in ["M", "A"]
            ]

            # untracked_files = [detection for detection in detections if detection.file_path in self.repo.untracked_files or detection.test.file_path in self.repo.untracked_files]
            # changed_files = [detection for detection in detections if detection.file_path in all_changes or detection.test.file_path in all_changes]

            # we must do this call BEFORE the list comprehension because otherwise untracked files are enumerated on each
            # iteration through the list and it is EXTREMELY slow
            repo_untracked_files = self.repo.untracked_files

            untracked_files = [
                detection
                for detection in detections
                if detection.file_path in repo_untracked_files
                or detection.test.file_path in repo_untracked_files
            ]
            changed_files = [
                detection
                for detection in detections
                if detection.file_path in all_changes_paths
                or detection.test.file_path in all_changes_paths
            ]
        except Exception as e:
            print(f"Error enumerating modified content: {str(e)}")
            sys.exit(1)

        return untracked_files, changed_files



from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingInfrastructure,
)
import docker.models.resource
import docker.models.containers
import docker
import docker.types
from contentctl.objects.test_config import (
    CONTAINER_APP_DIR,
    LOCAL_APP_DIR,
)


class DetectionTestingInfrastructureContainer(DetectionTestingInfrastructure):
    container: docker.models.resource.Model = None

    def start(self):
        self.container = self.make_container()
        self.container.start()

    def finish(self):
        if self.container is not None:
            try:
                self.removeContainer()
                pass
            except Exception as e:
                raise (Exception(f"Error removing container: {str(e)}"))
        super().finish()

    def get_name(self) -> str:
        return self.config.container_name

    def get_docker_client(self):
        try:
            c = docker.client.from_env()

            return c
        except Exception as e:
            raise (Exception(f"Failed to get docker client: {str(e)}"))

    def check_for_teardown(self):

        try:
            self.get_docker_client().containers.get(self.get_name())
        except Exception as e:
            if self.sync_obj.terminate is not True:
                self.pbar.write(
                    f"Error: could not get container [{self.get_name()}]: {str(e)}"
                )
                self.sync_obj.terminate = True

        if self.sync_obj.terminate:
            self.finish()

        super().check_for_teardown()

    def make_container(self) -> docker.models.resource.Model:
        # First, make sure that the container has been removed if it already existed
        self.removeContainer()

        ports_dict = {
            "8000/tcp": self.config.web_ui_port,
            "8088/tcp": self.config.hec_port,
            "8089/tcp": self.config.api_port,
        }

        mounts = [
            docker.types.Mount(
                source=str(LOCAL_APP_DIR.absolute()),
                target=str(CONTAINER_APP_DIR.absolute()),
                type="bind",
                read_only=True,
            )
        ]

        environment = {}
        environment["SPLUNK_START_ARGS"] = "--accept-license"
        environment["SPLUNK_PASSWORD"] = self.config.splunk_app_password
        environment["SPLUNK_APPS_URL"] = ",".join(
            p.environment_path for p in self.config.apps
        )
        if (
            self.config.splunkbase_password is not None
            and self.config.splunkbase_username is not None
        ):
            environment["SPLUNKBASE_USERNAME"] = self.config.splunkbase_username
            environment["SPLUNKBASE_PASSWORD"] = self.config.splunkbase_password

        container = self.get_docker_client().containers.create(
            self.config.full_image_path,
            ports=ports_dict,
            environment=environment,
            name=self.get_name(),
            mounts=mounts,
            detach=True,
        )

        return container

    def removeContainer(self, removeVolumes: bool = True, forceRemove: bool = True):

        try:
            container: docker.models.containers.Container = (
                self.get_docker_client().containers.get(self.get_name())
            )
        except Exception as e:
            # Container does not exist, no need to try and remove it
            return
        try:

            # container was found, so now we try to remove it
            # v also removes volumes linked to the container
            container.remove(v=removeVolumes, force=forceRemove)
            # remove it even if it is running. remove volumes as well
            # No need to print that the container has been removed, it is expected behavior

        except Exception as e:
            raise (
                Exception(
                    f"Could not remove Docker Container [{self.config.container_name}]: {str(e)}"
                )
            )




import shutil
import os

from dataclasses import dataclass


@dataclass(frozen=True)
class InitializeInputDto:
    path: str


class Initialize:

    def execute(self, input_dto: InitializeInputDto) -> None:

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/contentctl_default.yml'), 
            os.path.join(input_dto.path, 'contentctl.yml')
        )
           
        folders = ['detections', 'stories', 'lookups', 'macros', 'baselines', 'dist', 'tests', 'docs', 'reporting']
        for folder in folders:
            os.makedirs(os.path.join(input_dto.path, folder))

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/detections/anomalous_usage_of_7zip.yml'), 
            os.path.join(input_dto.path, 'detections', 'anomalous_usage_of_7zip.yml')
        )

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/tests/anomalous_usage_of_7zip.test.yml'), 
            os.path.join(input_dto.path, 'tests', 'anomalous_usage_of_7zip.test.yml')
        )

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/stories/cobalt_strike.yml'), 
            os.path.join(input_dto.path, 'stories', 'cobalt_strike.yml')
        )

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/macros/security_content_ctime.yml'), 
            os.path.join(input_dto.path, 'macros', 'security_content_ctime.yml')
        )

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), '../templates/macros/security_content_summariesonly.yml'), 
            os.path.join(input_dto.path, 'macros', 'security_content_summariesonly.yml')
        )

        print('The following folders were created: {0} under {1}.\nContent pack has been initialized, please run `new` to create new content.'.format(folders, input_dto.path))




import re
import sys

from pydantic import ValidationError

from contentctl.objects.story import Story
from contentctl.objects.enums import SecurityContentType
from contentctl.objects.config import Config
from contentctl.input.yml_reader import YmlReader


class StoryBuilder():
    story: Story

    def setObject(self, path: str) -> None:
        yml_dict = YmlReader.load_file(path)
        yml_dict["tags"]["name"] = yml_dict["name"]

        try:
            self.story = Story.parse_obj(yml_dict)
        except ValidationError as e:
            print('Validation Error for file ' + path)
            print(e)
            sys.exit(1)

    def reset(self) -> None:
        self.story = None

    def getObject(self) -> Story:
        return self.story

    def addDetections(self, detections: list, config: Config) -> None:
        matched_detection_names = []
        matched_detections = []
        mitre_attack_enrichments = []
        mitre_attack_tactics = set()
        datamodels = set()
        kill_chain_phases = set()

        for detection in detections:
            if detection:
                for detection_analytic_story in detection.tags.analytic_story:
                    if detection_analytic_story == self.story.name:
                        matched_detection_names.append(str(f'{config.build.splunk_app.prefix} - ' + detection.name + ' - Rule'))
                        mitre_attack_enrichments_list = []
                        if (detection.tags.mitre_attack_enrichments):
                            for attack in detection.tags.mitre_attack_enrichments:
                                mitre_attack_enrichments_list.append({"mitre_attack_technique": attack.mitre_attack_technique})
                        tags_obj = {"mitre_attack_enrichments": mitre_attack_enrichments_list}
                        matched_detections.append({
                            "name": detection.name,
                            "source": detection.source,
                            "type": detection.type,
                            "tags": tags_obj
                        })
                        datamodels.update(detection.datamodel)
                        if detection.tags.kill_chain_phases:
                            kill_chain_phases.update(detection.tags.kill_chain_phases)

                        if detection.tags.mitre_attack_enrichments:
                            for attack_enrichment in detection.tags.mitre_attack_enrichments:
                                mitre_attack_tactics.update(attack_enrichment.mitre_attack_tactics)
                                if attack_enrichment.mitre_attack_id not in [attack.mitre_attack_id for attack in mitre_attack_enrichments]:
                                    mitre_attack_enrichments.append(attack_enrichment)

        self.story.detection_names = matched_detection_names
        self.story.detections = matched_detections
        self.story.tags.datamodels = sorted(list(datamodels))
        self.story.tags.kill_chain_phases = sorted(list(kill_chain_phases))
        self.story.tags.mitre_attack_enrichments = mitre_attack_enrichments
        self.story.tags.mitre_attack_tactics = sorted(list(mitre_attack_tactics))


    def addBaselines(self, baselines: list) -> None:
        matched_baseline_names = []
        for baseline in baselines:
            for baseline_analytic_story in  baseline.tags.analytic_story:
                if baseline_analytic_story == self.story.name:
                    matched_baseline_names.append(str(f'ESCU - ' + baseline.name))

        self.story.baseline_names = matched_baseline_names

    def addInvestigations(self, investigations: list) -> None:
        matched_investigation_names = []
        matched_investigations = []
        for investigation in investigations:
            for investigation_analytic_story in  investigation.tags.analytic_story:
                if investigation_analytic_story == self.story.name:
                    matched_investigation_names.append(str(f'ESCU - ' + investigation.name + ' - Response Task'))
                    matched_investigations.append(investigation)

        self.story.investigation_names = matched_investigation_names
        self.story.investigations = matched_investigations

    def addAuthorCompanyName(self) -> None:
        match_author = re.search(r'^([^,]+)', self.story.author)
        if match_author is None:
            self.story.author_name = 'no'
        else:
            self.story.author_name = match_author.group(1)

        match_company = re.search(r',\s?(.*)$', self.story.author)
        if match_company is None:
            self.story.author_company = 'no'
        else:
            self.story.author_company = match_company.group(1)
